 /*
 * Program:	VMS mail routines
 *
 * Author:	Yehavi Bourvine, The Hebrew University of Jerusalem.
 *		Internet: Yehavi@VMS.huji.ac.il
 *
 * Modified:	Andy Harper, Kings College London
 *		Most of the original code has been rewritten but I've kept all
 *		acknowledgement notices anyway.
 *
 * Date:	2 August 1994
 * Last Edited:	18 July 1997
 *
 * Copyright 1993 by the University of Washington
 *
 *  Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appears in all copies and that both the
 * above copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the University of Washington not be
 * used in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  This software is made
 * available "as is", and
 * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
 * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
 * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <types.h>
#include <errno.h>
#include <signal.h>
#include "mail.h"
#include "osdep.h"
#include <sys/file.h>
#include <sys/stat.h>
#include "vms_mail.h"
#include "rfc822.h"
#include "misc.h"
#include "dummy.h"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <descrip.h>
#include <starlet.h>
#include <rms.h>
#include <fdldef.h>

int fdl$create ();
int fdl$generate ();

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

/*
  VMS debugging code
*/
#ifdef LOGTRACE
# define DBGPRINT(level,fmt,a,b,c,d) \
  {char *dbg = getenv("CCLIENT_LOGLEVEL"); \
   if (dbg && (atoi(dbg) >= level)) \
     {char tmpmsg[256];sprintf(tmpmsg,fmt,a,b,c,d); mm_dlog(tmpmsg); \
     } \
  }
#else
# define DBGPRINT(level,fmt,a,b,c,d) {}
#endif



/* CCLIENT DEBUG LEVELS */

#define DBG$ENTRY	21
#define DBG$EXIT	21
#define DBG$SYSSTATUS	23
#define DBG$SHOWMORE	25

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Miscellaneous bits and pieces
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

					/* Callable MAIL status codes */

#define MAIL$_NOTEXIST  0x007E80DA	/* MAIL - folder does not exist */
#define MAIL$_MSGTEXT   0x007EFD01	/* MAIL - text record returned */
#define MAIL$_NOMOREREC 0x007EDFA8	/* MAIL - No more records */

					/* Subscription database name */
#define MBXDATABASE ".MAILBOXLIST"


					/* Structures */

struct MailBoxList			/* For 'new' mail boxes */
  {
    struct MailBoxList *next;
    char *mailfile;
    char *folder;
  };

struct FolderPattern			/* For folder scanning */
  {
    int count;				/* Incremented if name matches */
    char *mailfile;			/* mailfile being scanned */
    char *prefix;			/* Prefix for folder pattern */
    char *pattern;			/* Folder name pattern */
  };

					/* Abbreviation for item lists */
typedef struct ITEM_LIST item;

					/* Global variables */
static struct MailBoxList *MailBoxes = NULL;
static item   NullItemList[] = {{0, 0, NULL, NULL}};





/* Driver dispatch used by MAIL */

DRIVER vmsmaildriver = {
  "vms_mail",			/* driver name */
  (DRIVER *) NIL,		/* next driver */
  vms_mail_valid,		/* mailbox is valid for us */
  vms_mail_parameters,		/* manipulate parameters */
  vms_mail_find,		/* find mailboxes */
  vms_mail_find_bboards,	/* find bboards */
  vms_mail_find_all,		/* find all mailboxes */
  vms_mail_find_all_bboards,	/* find all bboards */
  vms_mail_subscribe,		/* subscribe to mailbox */
  vms_mail_unsubscribe,		/* unsubscribe from mailbox */
  vms_mail_subscribe_bboard,	/* subscribe to bboard */
  vms_mail_unsubscribe_bboard,	/* unsubscribe from bboard */
  vms_mail_create,		/* create mailbox */
  vms_mail_delete,		/* delete mailbox */
  vms_mail_rename,		/* rename mailbox */
  vms_mail_open,		/* open mailbox */
  vms_mail_close,		/* close mailbox */
  vms_mail_fetchfast,		/* fetch message "fast" attributes */
  vms_mail_fetchflags,		/* fetch message flags */
  vms_mail_fetchstructure,	/* fetch message envelopes */
  vms_mail_fetchheader,		/* fetch message header only */
  vms_mail_fetchtext,		/* fetch message body only */
  vms_mail_fetchbody,		/* fetch message body section */
  vms_mail_setflag,		/* set message flag */
  vms_mail_clearflag,		/* clear message flag */
  vms_mail_search,		/* search for message based on criteria */
  vms_mail_ping,		/* ping mailbox to see if still alive */
  vms_mail_check,		/* check for new messages */
  vms_mail_expunge,		/* expunge deleted messages */
  vms_mail_copy,		/* copy messages to another mailbox */
  vms_mail_move,		/* move messages to another mailbox */
  vms_mail_append,		/* append string message to mailbox */
  vms_mail_gc			/* garbage collect stream */
};

				/* prototype stream */
MAILSTREAM vmsmailproto = {&vmsmaildriver};
                                /* standard driver's prototype */
MAILSTREAM *mailstd_proto = &vmsmailproto;

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	Simplify system status checking. Check the status message and print a
	message if it indicates an error. Exit with this status if the abort
        flag is set, otherwise return with LSB of the status.

    Input:
	message		message to display (often name of routine)
	abort		if T then abort, if NIL then just return
	status		status code to display

    Output:
	NONE

    Return Status:
	LSB of status code (1=ok, 0=error)

    Author:
 	Andy Harper, Kings College London, OCT 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int chkstat(char *message, int abort, int status)
  {
     DBGPRINT(DBG$SYSSTATUS,"%s: exit, status=0x%08x", message, status, 0, 0);
     if ((status & 0x1) == 0)
       {
          char errmess[256];
          sprintf(errmess, "%s: failed with status = 0x%08x", message, status);
          mm_dlog(errmess);

          if (abort)
            { fprintf(stderr,"%s\n", errmess); exit(status); }
       }

     return status & 0x1;
  }

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
      Case insensitive character match

   Inputs:
	c1		Character to check
	c2		Character to check

   Output:
	NONE

   Return Value:
	T		If a case-insensitive match
	NIL		If no match

   Author:
	Andy Harper, Kings College London, January 1997

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int chrcasecmp(char c1, char c2)
  {
     return (	(isupper(c1) ? tolower(c1) : c1) ==
		(isupper(c2) ? tolower(c2) : c2)  );
  }


/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Case insensitive string equality test

   Inputs:
	s1		String to check
	s2		String to check

   Output:
	NONE

   Return Value:
	T		If case-insensitive match
	NIL		If no match

   Author:
	Andy Harper, Kings College London, JAN 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int strcasecmp(char *s1, char *s2)
 {
    int n1 = strlen(s1);
    int n2 = strlen(s2);
    if (n1 != n2) return NIL;

    while (n1--)
     if ( !chrcasecmp(*s1++, *s2++) ) return NIL;

    return T;
 }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Match a case-insensitive wildcarded pattern

   Inputs:
	p		The pattern to match against
	s		The string to be matched

   Outputs:
	NONE

   Return Value:
	T		If a match
	NIL		If no match

   Comments:
	Special characters allowed in the pattern are:
		*	Match any string
		? or %	Match exactly one character

        Alphabetic characters are case-insensitive; other characters must match
	themselves exactly.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int matchwild(char *p, char *s)
  {
    int i;

    while (*p)
      switch (*p)
        {
         case '*' :				/* Wildcard match */
                    for (i=strlen(s); i>=0; i--)
                      if (matchwild(p+1,s+i)) return T;
                    return NIL;


         case '?':				/* Match any single char */
         case '%':
                    if (!*s++) return NIL; p++; break;


         default:				/* Case-Insensitive match */
                    if ( !chrcasecmp(*p++, *s++) ) return NIL;
        }

    return chrcasecmp(*p, *s);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Copy a string, uppercasing it as we go, except between quotes. The
        quotes themselves are removed.

   Inputs:
	in			The input string

   Outputs:
	out			The output string

   Return Value:
	NONE

   Author:
	Andy Harper, Feb 1997, Kings College London
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void __strucasecpy(char *out, char *in)
  {
    char c;

    while ( (c = *in++) )
      if (c == '"')
        {
          while ( (c = *in++) && (c != '"') )  *out++ = c;
          if ( c == '\0') break;
        }
      else
        *out++ =  islower(c) ? toupper(c) : c;

    *out = '\0';
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Create an empty file from the info contained in the FDL string

   Inputs:
	mailfile		Name of mailfile to be created
	fdl			FDL string
	fdlsize			Size of the FDL string

   Outputs:
	NONE

   Return Value:
	T			If file created OK
	NIL			If file not created

   Comments:

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __fdl_create_file(char *mailfile, char *fdl, int fdlsize)
  {
     int status, sts, stv;
     int flags   = FDL$M_FDL_STRING;
     struct dsc$descriptor
	fdl_desc = {fdlsize,          DSC$K_DTYPE_T, DSC$K_CLASS_S, fdl},
	filename = {strlen(mailfile), DSC$K_DTYPE_T, DSC$K_CLASS_S, mailfile};

     DBGPRINT(DBG$ENTRY,"__fdl_create_file: entered, mailfile='%s'",mailfile,0,0,0);
     status = fdl$create(&fdl_desc, &filename, NULL, NULL, NULL, &flags, NULL, NULL, &sts, &stv);
     if (!chkstat("fdl$create", 0, status))
       {
         chkstat("fdl$create sts", 0, sts);
         chkstat("fdl$create stv", 0, stv);
         return NIL;
       }

     DBGPRINT(DBG$EXIT,"__fdl_create_file: exit",0,0,0,0);
     return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Obtain an FDL description of a file into a string.

   Inputs:
	file			Name of file
        maxfdlsize		Size of output buffer

   Outputs:
	fdl			Returns the FDL string of the file
	fdlsize			Returns the string length

   Return Value:
	T			Obtained FDL ok
	NIL			FDL unobtainable

   Comments:
	In order to determine the key structure, we first get an XABSUM block
	which contains the number of keys and areas. We then use this to set up
	a dynamic XAB structure for this number of keys and areas.

	We use FDL$GENERATE to get details of the file. Note that the
	documentation on the output parameter is WRONG (as at time of writing)
	as it states that this must be a fixed-length string descriptor. It
	fact, it must be a dynamic string descriptor but with the buffer
	address pointer set to NULL. The buffer is allocated dynamically within
	FDL$GENERATE and must be copied and disposed of manually.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __fdl_generate(char *file, char *fdl, int maxfdlsize, int *fdlsize)
  {
    struct FAB    fab             = cc$rms_fab;
    struct RAB    rab             = cc$rms_rab;
    struct XABSUM xabsum          = cc$rms_xabsum;	/* XAB Summary info */
    struct XAB    *xabptr;

    int noa, nok;
    char keyname[25][32];			/* Space for 25 key names !! */
    struct XABKEY **xabkey;
    struct XABALL **xaball;

    int ret=NIL;				/* Default return status */
    int status;					/* VMS system status returns */


    DBGPRINT(DBG$ENTRY,"__fdl_generate: entered, file='%s'",file,0,0,0);
						/* Set up filename */
    fab.fab$l_fna = file;
    fab.fab$b_fns = strlen(file);

						/* Open file, get basic info */
    fab.fab$l_xab = &xabsum;
    status = sys$open(&fab);
    if (!chkstat("sys$open", 0, status)) return NIL;

						/* Set up RAB */
    rab.rab$l_fab = &fab;
    status = sys$connect(&rab);
    if (!chkstat("sys$connect", 0, status))
      {status=sys$close(&fab); return NIL;}

						/* Set up the area XABs ... */
    xaball = (struct XABALL **) &fab.fab$l_xab;
    for (noa=0; noa<xabsum.xab$b_noa; noa++)
      {
         *xaball = malloc(sizeof(struct XABALL));
         **xaball= cc$rms_xaball;
         (*xaball)->xab$b_aid = noa;		/* Area number */
         xaball  = (struct XABALL **) &(*xaball)->xab$l_nxt;
      }

						/* Set up the key XABs ... */
    xabkey = (struct XABKEY **) xaball;
    for (nok=0; nok<xabsum.xab$b_nok; nok++)
      {
         *xabkey = malloc(sizeof(struct XABKEY));
         **xabkey= cc$rms_xabkey;
         (*xabkey)->xab$b_ref = nok;		/* Key number */
	 (*xabkey)->xab$l_knm = keyname[nok];	/* Key name */
         xabkey  = (struct XABKEY **) &(*xabkey)->xab$l_nxt;
      }

						/* Re-fill all XAB blocks */
    status = sys$display(&fab);
    if (chkstat("sys$display",0,status))
      {
        struct dsc$descriptor FDLdesc;
        struct FAB *fabptr   = &fab;
        struct RAB *rabptr   = &rab;
        int FDLflags         = FDL$M_FDL_STRING;
						/* Parse FAB/RAB to get FDL */
        FDLdesc.dsc$w_length = maxfdlsize;
        FDLdesc.dsc$b_dtype  = DSC$K_DTYPE_T;
        FDLdesc.dsc$b_class  = DSC$K_CLASS_D;
        FDLdesc.dsc$a_pointer= NULL;
        status = fdl$generate(&FDLflags, &fabptr, &rabptr, NULL, NULL, &FDLdesc, NULL, fdlsize);
        if (chkstat("fdl$generate", 0, status))
          {
						/* Copy to the receive buffer */
            memcpy(fdl, FDLdesc.dsc$a_pointer, *fdlsize);
            fdl[*fdlsize] = '\0';
						/* Dispose of dynamic buffer */
            status = lib$sfree1_dd(&FDLdesc);
            chkstat("lib$sfree1_dd", 0, status);
            ret=T;
          }
      }

						/* Close the file */
    status = sys$close(&fab);
    chkstat("sys$close", 0, status);

						/* Dispose of area/key XAB blocks */
    xabptr = fab.fab$l_xab;
    while (xabptr)
      {
         struct XAB *ptr = xabptr;
         xabptr = xabptr->xab$l_nxt;
         free(ptr);
      }
        
    DBGPRINT(DBG$EXIT,"__fdl_generate: exit, status='%d'",ret,0,0,0);
    return ret;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Set message flags in the elt and the cache

   Inputs:
	stream		The mail stream for the mailbox
	i		The message number
	elt		Pointer to the elt for the message
	f		New flags to set

   Output:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __setflag(MAILSTREAM *stream, long i, MESSAGECACHE *elt, short f)
  {
      FILECACHE *fc = &LOCAL->msgs[i-1];
      DBGPRINT(DBG$ENTRY,"__setflag: entered, msgno='%ld', set = '0x%08x'", i, f, 0, 0); 
				/* set all requested flags */
      if (f&fSEEN)     {elt->seen     = T; fc->flags |= fSEEN;}
      if (f&fDELETED)  {elt->deleted  = T; fc->flags |= fDELETED;}
      if (f&fFLAGGED)  {elt->flagged  = T; fc->flags |= fFLAGGED;}
      if (f&fANSWERED) {elt->answered = T; fc->flags |= fANSWERED;}
      DBGPRINT(DBG$EXIT,"__setflag: exit", 0, 0, 0, 0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Clear message flags in the elt and the cache

   Inputs:
	stream		The mail stream for the mailbox
	i		The message number
	elt		Pointer to the elt for the message
	f		New flags to clear

   Output:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __clearflag(MAILSTREAM *stream, long i, MESSAGECACHE *elt, short f)
  {
      FILECACHE *fc = &LOCAL->msgs[i-1];
      DBGPRINT(DBG$ENTRY,"__clearflag: entered, msgno='%ld', clear = '0x%08x'", i, f, 0, 0); 
				/* clear all requested flags */
      if (f&fSEEN)     {elt->seen     = NIL; fc->flags &= ~fSEEN;}
      if (f&fDELETED)  {elt->deleted  = NIL; fc->flags &= ~fDELETED;}
      if (f&fFLAGGED)  {elt->flagged  = NIL; fc->flags &= ~fFLAGGED;}
      if (f&fANSWERED) {elt->answered = NIL; fc->flags &= ~fANSWERED;}
      DBGPRINT(DBG$EXIT,"__clearflag: exit", 0, 0, 0, 0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	Get the default MAIL directory for the current user

    Input:
	NONE

    Output:
	Mail		Pointer to mail directory buffer pointer; if NULL
			nothing returned here.

    Return Value:
	Mail directory buffer pointer

    Comments:
	Remember what we found on the first successfull call and don't call
	the system routines again if called again.

    Author:
	Andy Harper, Kings College London, OCT 1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

char *vms_get_mail_directory(char **Mail)
{
  static char MailDirectory[MAILFILESIZE];
  static int PrevCall = 0;

  DBGPRINT(DBG$ENTRY,"vms_get_mail_directory: entered",0,0,0,0);

  if (!PrevCall)
    {
      char username[80], host[256], password[80];
      int UserContext=0, length=0, status;

      item InItemList[] = {
	{ 0,              MAIL$_USER_USERNAME,       username,      NULL},
	{ 0, 0, NULL, NULL }};

      item OutItemList[] = {
	{ MAILFILESIZE-1, MAIL$_USER_FULL_DIRECTORY, MailDirectory, &length},
	{ 0, 0, NULL, NULL }};

					/* Ensure we get the right user */
      mm_login(host, username, password, -1L);
      InItemList[0].length = strlen(username);

					/* Quiz MAIL for the directory */
      status = mail$user_begin(&UserContext, NullItemList, NullItemList);
      if (!chkstat("mail$user_begin", 0, status) ) return NULL;
      status = mail$user_get_info(&UserContext, InItemList, OutItemList);
      if (!chkstat("mail$user_get_info", 0, status) ) return NULL;
      status = mail$user_end(&UserContext, NullItemList, NullItemList);
      if (!chkstat("mail$user_end", 0, status) ) return NULL;

					/* Terminate string, store pointer */
      MailDirectory[length] = '\0';
      PrevCall = 1;
    }
  if (Mail) *Mail = MailDirectory;

  DBGPRINT(DBG$EXIT,"vms_get_mail_directory: exited, directory='%s'",MailDirectory,0,0,0);
  return MailDirectory;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Purpose:
	Get the login directory (SYS$LOGIN) for the current user

   Input:
	NONE

   Output:
	hp		Pointer to home directory buffer pointer; if NULL
			nothing is returned here.

   Return Value:
	home directory buffer pointer

   Comments:
	Remember what we found on the first call and don't call the system
	routines again if called again.

   Author:
	Andy Harper, Kings College London, 4-OCT-1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *__get_home_directory(char **hp)
  {
     static char home[100], device[32], directory[64];
     static int PrevCall = 0;

     DBGPRINT(DBG$ENTRY,"__get_home_directory: entered",0,0,0,0);

     if (!PrevCall)
       {
         char host[256], username[256], password[256];
         int status;
         int devlen, dirlen;

         struct DESC usernameD;
         item UserInfo[] = {
	    {sizeof(device),       UAI$_DEFDEV,  device,     0},
	    {sizeof(directory),    UAI$_DEFDIR,  directory,  0},
            {0,                    0,            0,          0}};

					/* Ensure we get the right user */
         mm_login(host, username, password, -1L);
         usernameD.address = ucase( username);
         usernameD.length  = strlen(username);
         usernameD.type    = 0;

					/* Get directory from UAF */
         status = sys$getuai(0,0,&usernameD, UserInfo,0,0,0);
         if (!chkstat("sys$getuai", 0, status) ) return NULL;

					/* Convert to null terminated string */
         sprintf(home,"%.*s%.*s",device[0],device+1,directory[0],directory+1);
         PrevCall = 1;
       }
     if (hp) *hp = home;

     DBGPRINT(DBG$EXIT,"__get_home_directory: exit, directory='%s'",home,0,0,0);
     return home;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Purpose:
	To parse the supplied mailbox specification, supply defaults where
	necessary (mail directory etc.), and split into a mail file and
	folder specification.

   Input Formats:
        INBOX			--> [default mail file], NEWMAIL
	folder			--> [default mail file], folder

      The following for unix: compatibility:
	~/folder		--> [relative mail file], folder
        ~/subdir/folder		--> [relative mail file], folder

        The filename part is always "MAIL.MAI" and this should not be
	specified in the pathname.

        A NULL folder name is an error, any other format is invalid


   Input:
	name:		The mailbox name in standard format

   Outputs:
	mailfile	Pointer to VMS filename buffer, or NULL
	folder		Pointer to folder in mailfile buffer, or NULL
	inbox		Pointer to 'is an INBOX' flag

   Returns:
  	NIL mailbox parse error
	T   all ok

   Comments:
	The name of the 'new' mail folder defaults to "NEWMAIL" but can be
	overridden by the logical name CCLIENT_INBOXFOLDERNAME.

   Author:
	Andy Harper, Kings College London, 30-SEP-1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/


static int __parse_mailbox_name (char *name, char *mailfile, char *folder, int *inbox)
  {
    static char *DefaultInboxName = "NEWMAIL";		/* Standard 'new' mail folder */
    char mailbox[256], tmpmailfile[MAILFILESIZE];
    char *p, *InboxName, *mbox;

    if (!(InboxName=getenv("CCLIENT_INBOXFOLDERNAME"))) InboxName=DefaultInboxName;


/* -------------------Save Original mailbox Name ---------------------------- */

    mbox = mailbox;
    strcpy(mbox, name);

/* -------------------Are we looking at 'INBOX' ----------------------------- */

    if ( strcasecmp(mbox, "INBOX") )		/* Special INBOX name */
      {
         strcpy(tmpmailfile, vms_get_mail_directory(NULL) );
         mbox = InboxName;
      }

/* -------------------Relative mailbox specification ------------------------ */

    else if ( *mbox == '~' )			/* Home-relative specification */
      {
         if (mbox[1] != '/')			/* ... must start '~/'  ... */
           {
             DBGPRINT(DBG$EXIT,"__parse_mailbox_name: exit, status=0 : '/' must follow '~'",0,0,0,0);
             return NIL;
           }
         mbox += 2;

         strcpy(tmpmailfile,__get_home_directory(NULL));
         tmpmailfile[strlen(tmpmailfile)-1] = '\0';	/* Strip ']' from end */

						/* strip paths, convert to VMS subdirs */
         while ( (p=strchr(mbox, '/') ) != NULL)
           {
              *p = '\0';			/* Mask out '/'  */
              strcat(tmpmailfile,".");		/* Add subdir separator */
              strcat(tmpmailfile,mbox);		/* Add subdir */
              mbox = p+1;			/* start of next part */
              if (*mbox == '/')			/* Adjacent slashes = error */
                {
                  DBGPRINT(DBG$EXIT,"__parse_mailbox_name: exit, status=0 : '//' invalid",0,0,0,0);
                  return NIL;
                }
           }
         strcat(tmpmailfile,"]");			/* Terminate directory part ..*/
      }

/* ------------------Simple Folder Name ------------------------------------- */

    else					/* Simple folder name */
     {
        if (strchr(mbox,'/') != NULL)		/* ... no paths allowed */
         {
           DBGPRINT(DBG$EXIT,"__parse_mailbox_name: exit, status=0, absolute path not permitted",0,0,0,0);
           return NIL;
         }
        strcpy(tmpmailfile,vms_get_mail_directory(NULL));
     }


/* -------------------------------------------------------------------------- */

    if (*mbox == '\0')				/* missing folder name */
      {
        DBGPRINT(DBG$EXIT,"__parse_mailbox_name: exit, status=0 : missing folder name",0,0,0,0);
        return NIL;
      }


						/* Are we the INBOX ? */
    *inbox = strcasecmp(tmpmailfile, vms_get_mail_directory(NULL)) &&
            (strcmp    (mbox,        InboxName) == 0);
						/* Append filename */
    strcat(tmpmailfile,"MAIL.MAI");

						/* Copy, if requested */
    if (mailfile) strcpy(mailfile, tmpmailfile);
    if (folder)   strcpy(folder,   mbox);

    DBGPRINT(DBG$EXIT,"__parse_mailbox_name: exit, mailfile='%s', folder='%s', inbox='%d'",tmpmailfile, mbox, *inbox, 0);
    return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Update the new mail count by subtracting, from the current count, the
	number of messages seen

   Input:
	MessagesSeen		Number of messages seen

   Output:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __update_newmail_count(int MessagesSeen)
{
  static char username[256];
  static short Count;
  char password[256],host[256];
  int status, UserContext=0;

  item OldMessageUser[] = {
	{ 0,             MAIL$_USER_USERNAME,         &username, NULL},
	{ 0, 0, NULL, NULL }};

  item OldMessageCount[] = {
	{ sizeof(Count), MAIL$_USER_NEW_MESSAGES,     &Count,    NULL },
	{ 0, 0, NULL, NULL }};

  item NewMessageCount[] = {
	{ 0,             MAIL$_USER_USERNAME,         &username, NULL},
	{ sizeof(Count), MAIL$_USER_SET_NEW_MESSAGES, &Count,    NULL },
	{ 0, 0, NULL, NULL }};

  DBGPRINT(DBG$ENTRY, "__update_newmail_count: entered, Messages seen = '%d'", MessagesSeen, 0, 0, 0);

		/* Make sure we update the count for the current username */  
  mm_login(host, username, password, -1L);
  OldMessageUser[0].length = strlen(username);
  NewMessageCount[0].length = strlen(username);

		/* Get current count and update by specified amount */
  status = mail$user_begin(&UserContext, NullItemList, NullItemList);
  if (!chkstat("mail$user_begin", 0, status) ) return;

  status = mail$user_get_info(&UserContext, OldMessageUser, OldMessageCount);
  if (!chkstat("mail$user_get_info", 0, status) ) return;
  DBGPRINT(DBG$SHOWMORE,"__update_newmail_count: old count = '%d'", Count, 0, 0, 0); 
  if (MessagesSeen < Count)		/* Don't set a negative count */
    Count -= (short) MessagesSeen;
  else
    Count  = 0;
  DBGPRINT(DBG$SHOWMORE,"__update_newmail_count: new count = '%d'", Count, 0, 0, 0);
  status = mail$user_set_info(&UserContext, NewMessageCount, NullItemList);
  if (!chkstat("mail$user_set_info", 0, status) ) return;

  status = mail$user_end(&UserContext, NullItemList, NullItemList);
  if (!chkstat("mail$user_end", 0, status) ) return;

  DBGPRINT(DBG$EXIT, "__update_newmail_count: exit", 0, 0, 0, 0);
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Update the message status flags

   Inputs:
	Stream		mail stream
	msgno		Message to update
	syncflag	T   = synchronize external folder
			NIL = just update local flags.

   Output:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

static void __update_message_status (MAILSTREAM *stream, long msgno, long syncflag)
  {
     MESSAGECACHE *elt = mail_elt(stream,msgno);
     FILECACHE    *fc = &LOCAL->msgs[msgno-1];
     short sysflags = 0;
     int status;

     item MessageId[] = {
	{sizeof(msgno),    MAIL$_MESSAGE_ID,    &msgno,    NULL},
	{0,                0,                   NULL,      NULL}};

     item GetFlagsItem[] = {
	{sizeof(sysflags), MAIL$_MESSAGE_RETURN_FLAGS, &sysflags, NULL},
	{0,                0,                   NULL,      NULL}};

     item SetFlagsItem[] = {
	{sizeof(msgno),    MAIL$_MESSAGE_ID,    &msgno,    NULL},
	{sizeof(sysflags), MAIL$_MESSAGE_FLAGS, &sysflags, NULL},
	{0,                MAIL$_NOSIGNAL,      NULL,      NULL},
	{0,                0,                   NULL,      NULL}};

     DBGPRINT(DBG$ENTRY,"__update_message_status: entered, msgno='%d', syncflag='%d'", msgno, syncflag, 0, 0);
     DBGPRINT(DBG$SHOWMORE,"__update_message_status: seen='%d', deleted='%d', answered='%d',flagged='%d'", elt->seen, elt->deleted, elt->answered, elt->flagged);

					/* Update cache flags from elt */
     DBGPRINT(DBG$SHOWMORE,"__update_message_status: old cache flags = '0x%08x'",fc->flags, 0, 0, 0);

					/* if message deleted, mark it seen */
     if (elt->deleted)
       elt->seen = T;

					/* If message now seen but wasn't before, update new mail count */
     DBGPRINT(DBG$SHOWMORE,"__update_message_status: elt->seen='%d', fc->seen='%d'", elt->seen, (fc->flags)&fSEEN, 0, 0);
     if (elt->seen)
       { if ( syncflag && !((fc->flags)&fSEEN) ) __update_newmail_count(1);
         __setflag(stream, msgno, elt, fSEEN);
       }
     else
       __clearflag(stream, msgno, elt, fSEEN);

     if (elt->deleted)
       __setflag(stream, msgno, elt, fDELETED);
     else
       __clearflag(stream, msgno, elt, fDELETED);

     if (elt->flagged)
       __setflag(stream, msgno, elt, fFLAGGED);
     else
       __clearflag(stream, msgno, elt, fFLAGGED);

     if (elt->answered)
       __setflag(stream, msgno, elt, fANSWERED);
     else
       __clearflag(stream, msgno, elt, fANSWERED);

     DBGPRINT(DBG$SHOWMORE,"__update_message_status: new cache flags = '0x%08x'", fc->flags, 0, 0, 0);


					/* Update mailfile with message flags */
     if (syncflag)
       {
         DBGPRINT(DBG$SHOWMORE,"__update_message_status: syncing message %d flags with mailfile", msgno, 0, 0, 0);
         status = mail$message_get(&LOCAL->msg_context, MessageId, GetFlagsItem);
         if (!chkstat("mail$message_get", 0, status)) return;
         DBGPRINT(DBG$SHOWMORE,"__update_message_status: current sys flags = '0x%08x' (seen='%d', flagged='%d', answered='%d')", sysflags, elt->seen, elt->flagged, elt->answered);

         if (elt->seen)    sysflags &= ~MAIL$M_NEWMSG;   else sysflags |= MAIL$M_NEWMSG;
         DBGPRINT(DBG$SHOWMORE,"__update_message_status: seen='%d', new sysflags='0x%08x'", elt->seen, sysflags, 0, 0);
         if (elt->flagged) sysflags |= MAIL$M_MARKED;   else sysflags &= ~MAIL$M_MARKED;
         DBGPRINT(DBG$SHOWMORE,"__update_message_status: flag='%d', new sysflags='0x%08x'", elt->flagged, sysflags, 0, 0);
         if (elt->answered)sysflags |= MAIL$M_REPLIED;  else sysflags &= ~MAIL$M_REPLIED;
         DBGPRINT(DBG$SHOWMORE,"__update_message_status: answ='%d', new sysflags='0x%08x'", elt->answered, sysflags, 0, 0);

         DBGPRINT(DBG$SHOWMORE,"__update_message_status: new     sys flags = '0x%08x'", sysflags, 0, 0, 0);
         status = mail$message_modify(&LOCAL->msg_context, SetFlagsItem, NullItemList);
         if (!chkstat("mail$message_modify", 0, status)) return;
         DBGPRINT(DBG$SHOWMORE,"__update_message_status: message/mailfile sync complete",0,0,0,0);
       }

    DBGPRINT(DBG$EXIT,"__update_message_status: exit, message '%d' status updated",msgno,0,0,0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Determine the VMS MAIL transport prefix

   Input:
	NONE

   Output:
	NONE

   Return Value:
	Pointer to prefix name, or NULL

   Comments:
	The prefix is determined from two logical names:
	  *  PINE_MAIL_PROTOCOL
          *  MAIL$INTERNET_TRANSPORT

	The latter is usually defined for OpenVMS 6.2 and above

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *__transport_name(void)
  {
     char *p;

     if ((p=getenv("PINE_MAIL_PROTOCOL")))
       return p;

     else if ((p=getenv("MAIL$INTERNET_TRANSPORT")))
       return p;

     else
       return NULL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Find the first character in a string which matches any one of a set
	of characters (similar to the 'C' function strpbrk).

   Inputs:
	s		Pointer to the string to be searched
	size		length of string to be searched
	charset		Pointer NULL-terminated string of characters to find

   Output:
	NONE

   Return Value:
	Offset to the first matching character (starting at 1), or zero if
	not found

   Author:
	Andy Harper, Kings College London, March 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int strbreak(char *s, int size, char *charset)
  {
     char *p = s;

     DBGPRINT(DBG$ENTRY,"strbreak: searching '%.*s' for char in set '%s'", size, s, charset, 0);

     while (size--)
       if (strchr(charset, *p++) != NULL) return p-s;

     return 0;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Scan a string, defined by a start and a length, for a substring and
	return the starting point of the substring.

	Similar to 'strstr' but is case-insensitive.

   Inputs:
	s		pointer to string to be searched
	size		Size of string to be searched
	item		Pointer to null-terminated string to be found

   Output:
	NONE

   Return Value:
	Pointer to start of substring

   Author:
	Andy Harper, Kings College London, Feb 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *strfind(char *s, int size, char *item)
  {
    int i, n = strlen(item);
    int match;

    DBGPRINT(DBG$ENTRY, "strfind: scanning '%.*s' for '%s'", size, s, item, 0); 
    while (size-- >= n)
      {
         match = 1;

         for (i=0; i<n; i++)
           if (!chrcasecmp(s[i], item[i]))
              {match=0; break;}

         if (match)
           {
             DBGPRINT(DBG$EXIT,"strfind: string found : '%.*s'",n,s,0,0);
             return s;      
           }

         s++;
      }

    DBGPRINT(DBG$EXIT, "strfind: string not found",0,0,0,0);
    return NULL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Find the length of the next item assuming that the input is a list of
        items separated by one of a set of characters, properly skipping
        quoted strings and escaped single characters (" ... "  and \x).

   Inputs:
	address		Pointer to string containing zero or more items
	n		Length of string
	charset		null-terminated string of terminator characters

   Output:
	NONE

   Return Value:
	Length of the item

   Comments:
	The separators are normally specified as comma, space and tab

   Author:
	Andy Harper, Kings College London, Feb 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int stritem(char *address, int n, char *charset)
  {
    char *a = address;
    int quotenxt=0, inquotes=0;


    DBGPRINT(DBG$ENTRY,"stritem: input = '%.*s'", n, address, 0, 0);
    while (n--)
     {
       if (quotenxt)			/* Need to skip next char */
         quotenxt=0;

       else if (*a == '\\')		/* Quoting flag ?? */
         quotenxt = ~quotenxt;

       else if (*a == '"')		/* Start/end of quoted string */
         inquotes = ~inquotes;

					/* Unquoted separators */
       else if (!inquotes && (strchr(charset, *a) != NULL))
         break; 

       a++;
      }


     DBGPRINT(DBG$EXIT, "stritem: item='%.*s'", a-address, address, 0, 0);
     return a-address;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Ensure that the local (username) part of the smtp style address is
        appropriately quoted if it contains any special characters
        (which should already be individually quoted)

   Inputs:
	local		Local part (username) of the SMTP address string
	lsize		Length of the local part
	dom		domain name string
	dsize		length of domain name

   Outputs:
	output		Output buffer for rewritten string

   Function Return Value:
	length of string written to output buffer

   Comments:
	According to the RFC822 spec (which is admittedly a bit unclear),
	characters may be 'quoted' by preceding them with a '\' character. This
	may only be done inside a quoted string however which seems a little
	strange, since that implies double quoting.

	Strictly, spaces and tabs may appear unquoted within the string but any
	of the special characters should be prefixed by \.

	The addresses processed by this routine are assumed to have the
	original quoting retained, with only the VMS specific surrounding quotes
	removed.

	Some examples of valid RFC822 addresses, as output by this routine
	   Joe@org			--->	Joe@org
           "Joe Smith"@org		--->	"Joe Smith"@org
           Joe Smith@org		--->	"Joe Smith"@org
           Joe\ Smith@org		--->	"Joe\ Smith"@org
	   Joe \"Mad Dog\" Smith@org	--->    "Joe \"Mad Dog\" Smith"@org
	   clark::kent@org		--->	"clark::kent"@org

	Here are some tricky examples that need work:
	   "Joe"."Smith"@org		--->	"Joe.Smith"@org
				    or	--->	"\"Joe\".\"Smith\""@org


	Note that:
	   \"Mad Dog\"@org is NOT the same as "Mad Dog"@org !

	Note that this one is not a valid RFC822 address
	   Joe\ Smith@org		( must be inside "...")

	Additionally, some VMS mail transports may substitute ' for the
	enclosing " characters, to avoid confusion with the quotes used by
	the tsp%"... @ ..." format in VMS MAIL. So the following syntax may
	also appear:
	   'Joe Smith'@org		--->	"Joe Smith"@org
	   
   Author:
	Andy Harper, Kings College London, March 1997
	Claudio Allocchio, Sincrotrone Trieste, July 1997 (update)

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __rfc822_quote(char *output, char *local, int lsize, char *dom, int dsize)
  {
    static char rfc822_specials[] = "()<>@,;:\\\"[] \t";	/* not '.' */
    static char rfc822_specialq[] = "<>\\\"[]";
    char lhs[255];
    char *p;

    /* complex logic, but we cannot point outside 'local' string boundaries */

				/* Remove eventual external quotings */
    lhs[0] = '\0';

    if (lsize>=3)		/* "...." and no " inside string */
       if ((local[0]=='"') && (local[lsize-1]=='"') && (local[lsize-2]!='\\') &&
           ((p=strfind(local+1,lsize-2,"\""))==NULL)) 
          { strncpy(lhs, local+1, lsize-2); lhs[lsize-2] = '\0'; }


    if ((lhs[0]=='\0') && (lsize>=5))	/* \"....\" and no " inside string */
       if ((local[0]=='\\') && (local[1]=='"') && 
           (local[lsize-2]=='\\') && (local[lsize-1]=='"') &&
           ((p=strfind(local+2,lsize-4,"\""))==NULL))
          { strncpy(lhs, local+2, lsize-4); lhs[lsize-4] = '\0'; }


    if ((lhs[0]=='\0') && (lsize>=4))	/* "...." and \" only inside string */
       if ((local[0]=='"') && (local[lsize-1]=='"') && (local[lsize-2]!='\\'))
       {
          strncpy(lhs, local+1, lsize-2); lhs[lsize-2] = '\0';
          if (lhs[0] == '"')
             lhs[0] = '\0';
          p = lhs;
          while (*p != '\0')
          {
             if ((*p == '"') && (*(p-1) != '\\'))
                lhs[0] = '\0';
             p++;
          }
       }

   if (lhs[0] == '\0')
      { strncpy(lhs, local, lsize); lhs[lsize] = '\0'; }


				/*  Remove eventual already existing \x */
   p = lhs;
   while (*p != '\0')
   {
      if ((*p == '\\') && (strchr(rfc822_specials, *p) != NULL))
         memmove(p, p+1, strlen(lhs)-(p-lhs)+1);
      p++;
   }

				/*  Now, finally, quote RFC822 specials */
   p = lhs;
   while (*p != '\0')
   {
      if ((strchr(rfc822_specialq, *p) != NULL))
         { memmove(p+1, p, strlen(lhs)-(p-lhs)+2); *p = '\\'; p++; }
      p++;
   }


				/*  Contains any special RFC822 chars ?? */
    if (strbreak(lhs,strlen(lhs),rfc822_specials) != 0)
        return sprintf(output,"\"%s\"@%.*s", lhs, dsize, dom);


    else			/* Standard address - no quoting needed */
        return sprintf(output,"%s@%.*s", lhs, dsize, dom);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Convert a single VMS MAIL DECnet address into a conventional RFC822
        style of address.

   Inputs:
	address		Pointer to DECnet address to be converted
	n		Length of address string

   Output:
	output		Output buffer for converted string

   Return Value
	Length of converted string written to output

   Comments:
        Input formats are:
	  namespace:.nodename::user
	  nodename::user
          user

	Logical names are used to control the mapping.
	  *  CCLIENT_NODE[[_namespace]_nodename]
	  *  CCLIENT_NORFC1405
	  *  CCLIENT_GW_MAIL11

   Author:
	Andy Harper, Kings College London, Feb 1997
	Claudio Allocchio, Sincrotrone Trieste, July 1997 (update)

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __fmt_decnet(char *address, int size, char *output)
  {
    char *a            = address;
    int   n            = size;
    char *namespace    = "";
    int   namespacelen = 0;
    char *nodename     = "";
    int   nodenamelen  = 0;

    char logname[1024];
    char *sep, *domain;
    int nshort;

    DBGPRINT(DBG$ENTRY, "__fmt_decnet: address='%.*s'", size, address, 0, 0);


					/* Look for eventual first quotes char */
    if ((sep = strfind(a,n,"\"")) != NULL)
     nshort = sep-a; else nshort = n;


					/* Look for namespace (DECnet/OSI) */
    if ((sep = strfind(a,nshort,":.")) != NULL)
     {namespacelen = sep-a; namespace=a; n -= namespacelen+2; nshort -= namespacelen+2; a = sep+2;}


					/* Look for nodename (DECnet) */
    if ((sep = strfind(a,nshort,"::")) != NULL)
     {nodenamelen = sep-a; nodename=a; n -= nodenamelen+2; a = sep+2;}


					/* Construct the logical name */
    sprintf(logname, "CCLIENT_NODE%s%.*s%s%.*s",
          namespacelen>0 ? "_" : "", namespacelen, namespace,
          nodenamelen>0  ? "_" : "", nodenamelen,  nodename); 
    ucase(logname);


    DBGPRINT(DBG$SHOWMORE,"__fmt_decnet: address='%.*s', logical='%s'", n, a, logname, 0);


					/* Map the node via logical if possible */
    if ((domain=getenv(logname)) != NULL)
       return __rfc822_quote(output, a, n, domain, strlen(domain));


					/* Is non-RFC1405 mapping requested */
    else if (getenv("CCLIENT_NORFC1405") != NULL)
       return __rfc822_quote(output, a, n, nodename, nodenamelen);


					/* RFC1405 format with a specified gateway */
    else if ((domain=getenv("CCLIENT_GW_MAIL11")) !=NULL)
       return __rfc822_quote(output, address, size, domain, strlen(domain));


					/* RFC1405 format, no gateway */
    else if ((domain=mylocalhost()) !=NULL)
       return __rfc822_quote(output, address, size, domain, strlen(domain));


					/* RFC1405, no gateway, no hostname */
    else
       return __rfc822_quote(output, address, size, nodename, nodenamelen);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Convert a single VMS MAIL smtp style address into a conventional RFC822
        style of address.

   Inputs:
	address		Pointer to DECnet address to be converted
	n		Length of address string

   Output:
	output		Output buffer for converted string

   Return Value
	Length of converted string written to output

   Comments:
	Input formats accepted are:
	   node::"... @ ..."		--->	... @ ...
	   tsp%"... @ ..."		--->    ... @ ...
           "... @ ..."			--->    ... @ ...
           ... @ ...			--->	... @ ...

	The address can additionally contain either quoted quotes (\") or
	single quotes ('), which are converted to (") if found.

   Author:
	Andy Harper, Kings College London, Feb 1997
	Claudio Allocchio, Sincrotrone Trieste, July 1997 (update)

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __fmt_smtp(char *address, int n, char *output)
  {
    char *p;

    DBGPRINT(DBG$ENTRY,"__fmt_smtp: old address='%.*s'", n, address, 0, 0);


		/*   node::"..."    :: Remove nodename prefix */
    if ((p=strfind(address,n,"::\"")) != NULL)
      {n -= (p-address)+2; address = p+2;}


		/*   tsp%"....."    :: Remove transport name prefix */
    if ((p=strfind(address,n,"%")) != NULL)
      {n -= (p-address)+1; address = p+1;}


		/*   "....."        :: Remove enclosing double quotes */
    if ((*address == '"') && (address[n-1] == '"'))
      {address++; n--; n--;}

                /* '...'            :: Replace quoted single quotes */
    while (p=strfind(address,n,"'"))
      *p = '"';


		/*		    :: Quote rfc822 specials if needed */
    if ((p=strfind(address,n,"@")) != NULL)
      return __rfc822_quote(output,address,p-address,p+1,n-(p-address)-1);
    else	/* should never happen */
      return __rfc822_quote(output,address,n,mylocalhost(),strlen(mylocalhost()));
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Replace X.400 numeric and long keywords with standard short alpha
        keywords.

   Inputs:
	ps		pointer to 'keyword=value' string
	lps		Size of string
	n		Size of complete address

   Output:
	ps		pointer to new 'keyword=value' string
	lps		new size of string
	n		new size of complete address
	

   Return Value:
	NONE

   Author:
	Claudio Allocchio, Sincrotrone Trieste, Jul 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __fmt_x400_keys(char *ps, int lps, int nb)
  {
    char *p;
    int lpi;

    lpi = 0;

    if ((*ps == '*') && ((p=strfind(ps,lps,"\\")) != NULL))
      { *p = '='; memmove(ps+2,ps,nb); memcpy(ps,"DD.",3); lpi = 2; }
    else if (strfind(ps,lps,"1=") == ps)
      *ps = 'C';
    else if (strfind(ps,lps,"2=") == ps)
      *ps = 'A';
    else if (strfind(ps,lps,"3=") == ps)
      *ps = 'P';
    else if (strfind(ps,lps,"4=") == ps)
      { memmove(ps+1,ps,nb); memcpy(ps,"OU",2); lpi = 1; }
    else if (strfind(ps,lps,"5=") == ps)
      *ps = 'O';
    else if (strfind(ps,lps,"6=") == ps)
      *ps = 'S';
    else if (strfind(ps,lps,"7=") == ps)
      *ps = 'G';
    else if (strfind(ps,lps,"8=") == ps)
      *ps = 'I';
    else if (strfind(ps,lps,"9=") == ps)
      { memmove(ps+1,ps,nb); memcpy(ps,"GQ",2); lpi = 1; }
    else if (strfind(ps,lps,"10=") == ps)
      *ps = 'U';
    else if (strfind(ps,lps,"11=") == ps)
      *ps = 'X';
    else if (strfind(ps,lps,"12=") == ps)
      *ps = 'T';
    else if (strfind(ps,lps,"NAME=") == ps)
      { memmove(ps,ps+5,nb-5); lpi = -5; }

    return lpi;
  }
/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Convert a single VMS MAIL MAILBUS/MRGATE address into a conventional
	RFC822 style of address.

   Inputs:
	address		Pointer to address to be converted
	n		Length of address string

   Output:
	output		Output buffer for converted string

   Return Value
	Length of converted string written to output

   Comments:
	Input format is:
	  a=abc::m=def::z=ghi ... 

	Output format is:
	  "/a=abc/m=def/z=ghi ... /" @x400-gw

	Logical names used to control mapping:
	  * CCLIENT_GW_X400

   Author:
	Andy Harper, Kings College London, Feb 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __fmt_x400(char *address, int n, char *output)
  {
    char *p, *pd, *ps, *pg, *pi;
    int lps, lpg, lpi;
    char tmp[256];

    DBGPRINT(DBG$ENTRY, "__fmt_x400: address = '%.*s'", n, address, 0, 0);


					/* Save X.400 gw pointer */
    if ((pd=getenv("CCLIENT_GW_X400")) == NULL)
      strcpy(pd,"X400-GW");

					/* Replace '::' by '/'     */
    ps = address;			/* and non standard labels */
    pg = NULL; lpg=0;			/* pointer to Personal Name */
    if ((p=strfind(address,n,"::")) == NULL)
      lps = n;
    else
      lps = p-ps;
    lpi = __fmt_x400_keys(ps,lps,n+1);
    lps=lps+lpi; n=n+lpi; if ( p != NULL ) p=p+lpi; 
    if (strfind(ps,lps,"=") == NULL)
      { pg=ps; lpg=lps; }
    while ( p )
      { *p++ = '/'; memmove(p,p+1,n-(p-address)); n--; 
        ps = p;
        if ((p=strfind(address,n,"::")) == NULL)
          lps = n-(ps-address);
        else
          lps = p-ps;
        lpi = __fmt_x400_keys(ps,lps,n-(ps-address)+1);
        lps=lps+lpi; n=n+lpi; if ( p != NULL ) p=p+lpi;
        if (strfind(ps,lps,"=") == NULL)
          { pg=ps; lpg=lps; }
      }

					/* move Personal name as last elem. */
    if ((pg != NULL) && (pg != ps))
      {
        strncpy(tmp,pg-1,lpg+1);
        memmove(pg-1,pg+lpg,n-(pg+lpg-address));
        memcpy(address+n-lpg-1,tmp,lpg+1);
        ps=address+n-lpg; lps=lpg;
      }
    *(address+n)='\0';
					/* No Personal Name, return address */
    if ((p=strfind(ps,lps,"=")) != NULL)
      return sprintf(output,"\"\/%.*s\/\"@%s", n, address, pd);

    else
					/* Personal Name: Surname only */
      if ((p=strfind(ps,lps," ")) == NULL)
        { 
          n=n+2;
          return sprintf(output,"\"\/%.*sS=%.*s\/\"@%s", n-lps-2, address, lps, ps, pd);
        }

      else
					/* Personal Name: Givenname exists */
        {
          pg = ps;
          lpg = p-pg;
          ps = p+1;
          lps = lps-lpg-1;
					/* Personal Name: Surname, no Initl's */
          if ((p=strfind(ps,lps," ")) == NULL)
            { 
              n=n+4;
              return sprintf(output,"\"\/%.*sG=%.*s\/S=%.*s\/\"@%s", n-lpg-lps-5, address, lpg, pg, lps, ps, pd);
            }
          else
					/* Personal Name: Surname & Initl's */
            {
              pi = ps;
              lpi = p-pi;
              ps = p+1;
              lps = lps-lpi-1;
              n=n+6;
              return sprintf(output,"\"\/%.*sG=%.*s\/I=%.*s\/S=%.*s\/\"@%s", n-lpg-lpi-lps-8, address, lpg, pg, lpi, pi, lps, ps, pd);
            }
        }
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Convert a single VMS MAIL address into an RFC822 style address. This
	routine knows about the various formats that may appear in a VMS MAIL
	header.

   Inputs:
	address		Pointer to the original VMS MAIL address string
	n		Length of the address string

   Output:
	rfc822		Destination for the converted string

   Return Value:
	Length of string written to rfc822

   Comments:
	Leading and trailing white space is stripped first (though this should
	be unnecessary now, leave it in as a bit of defensive programming).

	if  MRGATE::"..." format
          then
           if X.400 (contains C=, 1= or COUNTRY=)
             then
	        remove MRGATE:: and enclosing "
                "x=abc::y=def:z=ghj"	-->	"/x=abc/y=def/z=ghj/"@x400-gw
             else
		remove enclosing quotes
                treat like DECnet address
                                          

        then, if contains '@'
           treat as an smtp address

	else, treat as a DECnet format 

   Author:
	Andy Harper, Kings College London, Feb 1997

   Acknowledgments:
	Claudio Allochio @ Luigi Fonti, INFN Italy

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __to_rfc822(char *address, int n, char *output)
  {
    char *p,*q;

    DBGPRINT(DBG$ENTRY,"__to_rfc822: size='%d', address='%.*s'", n, n, address,0);

					/* Skip leading white space */
    while ( (n>0) && isspace(*address))
      { address++; n--; }

					/* Skip trailing white space */
    while ( (n>0) && isspace(address[n-1]))
      n--;

					/* Misc. MRGATE formats ?? */
    if ((p=strfind(address, n, "MRGATE::\"")) != NULL)
      {
					/* X.400 or strange MAILBUS thing ?? */
         if ( (q=strfind(address,n,"c=")) || (q=strfind(address,n,"country=")) ||
              (q=strfind(address,n,"1=")) )
            {
					/* Skip 'MRGATE::', remove outer quotes */
              n -= (p-address) + 9 + 1;
              address = p + 9;
              return __fmt_x400(q, n-(q-address), output);
            }
         else				/* MRGATE:: but not X.400 */
            return __fmt_decnet(address, n, output);
      }

					/* SMTP ?? */
    else if ( strfind(address, n, "@") )
      return __fmt_smtp(address,n,output);


					/* Assume DECnet ?? */
    else
      return __fmt_decnet(address,n,output);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Extract each comma-separated address and convert it to a more
	conventional RFC822 format. Spaces are allowed as an alternative
        separator. Optional spaces are allowed around a comma separator.
        Spaces or commas within quoted strings are not separators.

   Inputs:
	record		Pointer to the address string(s)
	n		Length of the address string

   Output:
	record		The strings are rewritten in place

   Return Value:
	Length of new address field

   Comments:
	If at the start of an address, the line length exceeds 72 characters,
	then we insert a continuation line in accordance with the RFC
	specification. A \n is added to terminate the complete list.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __rewrite_address(char *record, int n)
 {
   char fieldcontent[256];
   char *a       =  fieldcontent;
   char *p       =  record;
   char *nl      =  record;
   char *sepstrng= "";			/* Initial separator string */
   char *pst;
   int   i;
					/* Make copy of address(es) */
   strncpy(fieldcontent, record, n);

					/* Loop through address list */
   while (n > 0)
    {
					/* Skip leading separators */
      while (n && ((*a == ',') || isspace(*a)))
        { a++; n--;}

					/* Find end of address string */
      if ((i=stritem(a,n,", \t")) > 0)
        {
					/* Add the current separator string */
          strcpy(p, sepstrng); p += strlen(sepstrng);
          sepstrng = ",  ";


          if ( (p-nl) > 72)		/* Break long line */
            { *p++ = '\n'; nl = p; *p++ = ' '; *p++ =' ';}


					/* Rewrite this address */
          DBGPRINT(DBG$SHOWMORE,"__rewrite_address: original ='%.*s'",i,a,0,0);
          pst=p;
          p += __to_rfc822(a, i, p);
          DBGPRINT(DBG$SHOWMORE,"__rewrite_address: rewritten='%.*s'",p-pst,pst,0,0);

          a += i;			/* Skip past old address */
          n -= i;
        }       
    }

   return p-record;			/* return new size */
 }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Rewrite standard VMS MAIL headers into a more standard RFC822 form

   Inputs:
	field		The mail header to be modified
	n		Length of header

   Output:
	field		Modified in situ as necessary

   Return Value:
	Length of rewritten address

   Comments:
	A FROM: field has additional info (personal name and date) that we
	don't want, so this must be eliminated.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __rewrite_vms_headers(char *field, int n)
  {
     static char subj[]="subj:", subject[]="Subject:";

     char *content;
     int   contentlen, fieldlen;

					/* Find end of field name */
     content    = strfind(field, n, ":") + 1;
     *content++ = ' ';			/* Ensure space (not tab) terminator */
     fieldlen   = content - field;
     contentlen = n - fieldlen;


					/* Check which field we're processing*/
     if (strfind(field, fieldlen, "from:"))
       {
         contentlen = stritem(content,contentlen,", \t");/* Ignore personal name/date fields */
         content += __rewrite_address(content, contentlen);
       }

     else if (strfind(field, fieldlen, "to:") )
         content += __rewrite_address(content, contentlen);

     else if (strfind(field, fieldlen, "cc:"))
         content += __rewrite_address(content, contentlen);

     else if (strfind(field, fieldlen, "subj:"))
       {
         memmove(field+strlen(subject), field+strlen(subj), n-strlen(subj));
         memmove(field,subject,strlen(subject));
         content += contentlen + (strlen(subject) - strlen(subj));
       }

					/* Add trailing \n, return size */
     *content ++ = '\n';
     return content-field;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	This routine examines the header record and returns T if it looks like
	an RFC822 header and NIL otherwise. Address fields must contain an
	'@' symbol to reduce the risk of lines starting with a valid header
        being mistakenly interpreted.

    Inputs:
	record		Header record buffer
	size		Size of the line in the buffer

    Output:
	NONE

    Return Value:
	T		Record appears to be an RFC822 mail header
	NIL		Record is not a header

    Author:
	Andy Harper, Kings College London, Sep 30 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __is_rfc822_header(char *record, int size)
  {
    static char *prefix[] =
	{"x400-","resent-","original-","originally-","original_"};
    static int prefixsize = sizeof(prefix) / sizeof(prefix[0]);

    static char *table[] =
	{"approved",
	 "autoforwarded",
	 "autosubmitted",
	 "bcc",
	 "cc",
	 "comment",
	 "comments",
	 "content-identifier",
	 "content-language",
	 "content-transfer-encoding",
	 "content-type",
	 "control",
	 "conversion-with-loss",
	 "conversion",
	 "date",
	 "delivery-date",
	 "discarded-ipms-extensions",
	 "discarded-mts-extensions",
	 "distribution",
	 "dl-expansion-deferred-delivery",
	 "encrypted",
	 "expires",
	 "followup-to",
	 "from",
	 "importance",
	 "in-reply-to",
	 "incomplete-latest-delivery-time",
	 "keywords",
	 "lines",
	 "message-id",
	 "message-original-encoded-information-types",
	 "mime-version",
	 "mts-identifier",
	 "newsgroups",
	 "organization",
	 "originator-return-address",
	 "originator",
	 "path",
	 "priority",
	 "read-receipt-to",
	 "received",
	 "recipients",
	 "references",
	 "relay-version",
	 "reply-by",
	 "reply-to",
	 "return-path",
	 "sender",
	 "sensitivity",
	 "subject",
	 "supsersedes",
	 "to",
	 "via:"};

    static int tablesize = sizeof(table) / sizeof(table[0]);

    int n;


  /* Skip over common prefix-headers */
    for (n=0; n<prefixsize; n++)
       if (strfind(record, size, prefix[n]) == record)
          {record += strlen(prefix[n]);  size -= strlen(prefix[n]); break;}


  /* Match against a standard RFC822 header at the start of the record */
    for (n=0; n<tablesize; n++)
      if (strfind(record, size, table[n]) == record)
         {return T;}


  /* Looks like an approved extension (X- ... :)? */
    if (    ( (n=strbreak(record,size,": \t")) > 2)
         && ( record[n-1] == ':' )
         && ( strfind(record, 2, "X-") == record ))       return T;


  /* Matches nothing, so not a header */
    return NIL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	This routine examines the header record key and sets the bit flags
	into hdrFields.

    Inputs:
	record		Header record buffer
	size		Size of the line in the buffer
	field		the RFC822 heade field
	fs		Size of RFC822 field
	fmask		Bitmask to be used
	hdrFields	original value of hdrFields bit flags

    Output:
	NONE

    Return Value:
	hdrFields	modified value of hdrFields bit flags

    Author:
	Claudio Allocchio, Sincrotrone Trieste, Jul 1997
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static unsigned short __set_headerSpecial(char *record, int size, char *field, int sf, unsigned short fmask, unsigned short hdrFields)
  {
    if ((strfind(record+sf,size-sf,":\tNAME:") == record+sf) ||
        (strfind(record+sf,size-sf,":\tFUNC:") == record+sf) ||
        (strfind(record+sf,size-sf,":\tINITLS:") == record+sf) ||
        (strfind(record+sf,size-sf,":     See Below") == record+sf))
      {
        hdrFields = hdrFields|fISMR;
        memcpy(record,"X400-",5);
        memcpy(record+5,field,sf);
        memcpy(record+5+sf,": ",2);
      }
    else if ((strfind(record,size,"@C=") != NULL) ||
             (strfind(record,size,"@A=") != NULL) ||
             (strfind(record,size,"@P=") != NULL) ||
             (strfind(record,size,"DD.RFC-822") != NULL))
      {
        hdrFields = hdrFields|fISMR;
        *record = 'X';
        memcpy(record+1,field,sf);
        *(record+1+sf) = ':';
      }
    else
      hdrFields = hdrFields|fmask;

    return hdrFields;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    Purpose:
	This routine examines the header record key and sets the bit flags
	into hdrFields.

    Inputs:
	record		Header record buffer
	size		Size of the line in the buffer
	hdrFields	original value of hdrFields bit flags

    Output:
	NONE

    Return Value:
	hdrFields	modified value of hdrFields bit flags

    Author:
	Claudio Allocchio, Sincrotrone Trieste, Jul 1997
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static unsigned short __set_headerFields(char *record, int size, unsigned short hdrFields)
  {
    if (strfind(record,size,"Sender") == record)
      hdrFields = hdrFields|fSENDER;
    else if (strfind(record,size,"Subject") == record)
      hdrFields = hdrFields|fSUBJECT;
    else if (strfind(record,size,"From") == record)
      hdrFields = __set_headerSpecial(record,size,"From",4,fFROM,hdrFields);
    else if (strfind(record,size,"To") == record)
      hdrFields = __set_headerSpecial(record,size,"To",2,fTO,hdrFields);
    else if (strfind(record,size,"Cc") == record)
      hdrFields = __set_headerSpecial(record,size,"CC",2,fCC,hdrFields);
    else if (strfind(record,size,"Bcc") == record)
      hdrFields = __set_headerSpecial(record,size,"BCC",3,fBCC,hdrFields);

			/* return new hdrFields value */
    return hdrFields;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Get information about a mail message

   Inputs:
	stream		Mail stream
	msgno		Number of message for which info required

   Outputs:
	size		Pointer to maximum byte size of message
	flags		Pointer to message status flags
	arrivaldate	Pointer to arrival date string pointer

   Return Value:
        NONE

   Comments:
	The 'size' is returned as a record count so we convert it a maximum
	byte count by multiplying by the maximum record count. In practice,
	records will be much less than this byte count but we wont ever go
	bigger so allocating this much buffer space later is pretty safe (if a
	little excessive). We can't find the actual byte count without reading
	the whole message.

	The meaning of the message flag bits are as follows (taken from the
	maildef.h include header):

	BIT	DECIMAL	SYMBOLIC NAME		MEANING
	  0	1	MAIL$M_NEWMSG		A NEW message
	  1	2	MAIL$M_REPLIED		Message has been REPLIED to
	  2	4	MAIL$M_DEL		Message has been deleted
	  3	8	MAIL$M_EXTMSG		Message text is in external file
	  4	16	 :
	  5	32	 :  Miscellaneous status, not useful to us
	  6	64	 :
	  7	128	MAIL$M_MARKED		Message has been marked


   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __message_info(MAILSTREAM *stream, long msgno, unsigned long *size, short *flags, char **arrivaldate)
  {
    static char DateBuff[256];
    static unsigned int RecCount, DateLen=0;
    int status;

    item MessageIdItem[] = {
	{sizeof(msgno),       MAIL$_MESSAGE_ID,           &msgno,    NULL },
	{0, 0, NULL, NULL }};

    item MessageInfoItem[] = {
	{sizeof(RecCount),    MAIL$_MESSAGE_SIZE,         &RecCount, NULL },
	{sizeof(DateBuff)-1,  MAIL$_MESSAGE_DATE,         DateBuff,  &DateLen},
	{sizeof(*flags),      MAIL$_MESSAGE_RETURN_FLAGS, flags,     NULL },
	{0, 0, NULL, NULL }};

  DBGPRINT(DBG$ENTRY,"__message_info: entered, mailbox='%s', msgno='%ld'", stream->mailbox, msgno,0,0);


					/* Obtain message info */
  status = mail$message_get(&LOCAL->msg_context, MessageIdItem, MessageInfoItem);
  chkstat("mail$message_get", 1, status);


					/* Convert/copy message info */
  *size = (RecCount+4)*MAXRECORDSIZE;	/* Allow 4 extra for VMS MAIL headers */
  *arrivaldate = DateBuff;
  DateBuff[DateLen] = '\0';

  DBGPRINT(DBG$EXIT, "__message_info: exit, records='%u', size='%lu', flags='0x%08x', date='%s'", RecCount, *size, *flags, DateBuff);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Obtain the next record of the current message from the MAIL folder

   Input:
	stream		The mail stream of the folder

   Output:
	record		Pointer to text buffer for the record

   Return Value:
	record length (including newline), 0 if EOF, -1 if error

   Comments:
	The message context should have been previously selected.

	The record is terminated by a newline character

   Author:
	Andy Harper, Kings College London, 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

static int __message_readline(MAILSTREAM *stream, char *record)
  {
    static int RecordLength;
    int i, status;

    static item GetMessageTextItem[] = {
        { sizeof(int),   MAIL$_MESSAGE_CONTINUE, NULL, NULL },
        { 0, 0, NULL, NULL }};

    static item RecordItem[] = {
        { MAXRECORDSIZE, MAIL$_MESSAGE_RECORD,   NULL, &RecordLength },
        { 0, 0, NULL, NULL }};


					/* Get next record of message */
    RecordItem[0].buffer = record;
    status = mail$message_get(&LOCAL->msg_context, GetMessageTextItem, RecordItem);

    if (status == MAIL$_MSGTEXT)	/* Found record OK  */
      { record[RecordLength]='\n'; return RecordLength+1; }

    if (status == MAIL$_NOMOREREC)	/* End of message */
      { return 0; }

					/* Some unknown error !!! */
    chkstat("mail$message_get",0,status);
    DBGPRINT(DBG$SHOWMORE,"__message_readline: unexpected error '0x%08x' reading message",status,0,0,0);
    return -1;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  Purpose:
    The message is read, line by line, from the mail file and, under VMS, has
    this format:

         VMS MAIL headers (From, To, CC, Subj)
         Blank Line
         [Optional RFC822 headers (if delivered by a network transport)
           or
          Optional X.400 P2 headers (if delivered via MRGATE)
          Blank line]
         [Optional RFC822 additional headers (if delivered from X.400 via MIXER)
          <FF>
          Blank line]
         Body of text

    This routine reads the message and separates it into headers
    and text parts:

    - where RFC822 headers are (apparently) present and valid, they overwrite 
    the VMS MAIL headers.

    - where X.400 P2 headers are present, they are appended to VMS Mail headers
    and reformatted into an rfc822-like syntax where possible.

    - where the additional RFC822 headers are present, they are appended to
    VMS Mail headers and X.400 P2 headers.

    - otherwise the VMS MAIL headers are reformatted into an rfc822-like syntax 
    where possible.

    As a result, the routine can decode nearly any MIME-like message format
    existing in VMS mailboxes.

   Inputs:
	stream		The mailstream
	msgno		The message to read

   Outputs:
	NONE

   Return Value:
	NONE

   Comments:
	This routine takes care to avoid the use of the C null string
	terminator where possible in order to allow null bytes to appear in a
	message (this may not be usual but it's a restriction that should not
	be imposed, so we don't).

	We use the LOCAL->buf structure as a temporary buffer in which to read
	the message. It is resized if needed.

	The logical name CCLIENT_HEADERS_ATEND specifies that the RFC822 headers
	have been placed at the end of the message by the mail delivery agent
	and defines the line which delimits the text from the headers. The
	line must be defined IN FULL and IN THE CORRECT CASE.

        The message is cached on the first access; subsequent reads do nothing

   Author:
	Andy Harper, Kings College London, Jan 1997
	Claudio Allocchio, Sincrotrone Trieste, July 1997 (update)

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void __read_message (MAILSTREAM *stream, long msgno)
  {
     FILECACHE     *fc = &LOCAL->msgs[msgno-1];
     MESSAGECACHE  *elt;
     char          *mbuf, *mbptr, *vmsstarthdr, *starthdr, *starttext;
     char          *deliverydate, *headstr, *rfcstarthdr;
     unsigned long size, vmsheadsize, headsize, bodysize, headstrlen, rfcheadsize;
     int           len;
     short         flags;
     unsigned short hdrFields;

     DBGPRINT(DBG$ENTRY,"__read_message: entered, mailbox='%s', msgno='%ld'",stream->mailbox, msgno,0,0);

						/* If cached, nothing to do */
     if (fc->text && fc->header)
       {
         DBGPRINT(DBG$EXIT,"__read_message: exit, message in cache",0,0,0,0);
         return;
       }


  /* ---------------- Get basic info about message ----------------- */

     elt = mail_elt(stream,msgno);
     __message_info(stream, msgno, &size, &flags, &deliverydate);

     elt->seen     = ((flags & MAIL$M_NEWMSG)  ? NIL : T);
     elt->answered = ((flags & MAIL$M_REPLIED) ? T : NIL);
     elt->flagged  = ((flags & MAIL$M_MARKED)  ? T : NIL);

     __update_message_status(stream,msgno,NIL);


  /* ----------------  Local buffer setup and initialization ------ */
						/* Resize buffer if needed */
     if (size >= LOCAL->buflen)
       {
         fs_give((void **) &LOCAL->buf);
         LOCAL->buf = (char *) fs_get((LOCAL->buflen=size)+1);
       }

						/* Initialize buffer pointers */
     mbptr = mbuf  = LOCAL->buf;

						/* Header placement */
     headstrlen = (headstr=getenv("CCLIENT_HEADERS_ATEND")) ? strlen(headstr) : 0;
     DBGPRINT(DBG$SHOWMORE,"__read_message: 'at end' string = '%s'",headstr?headstr:"(null)",0,0,0);

  /* -----------------   VMS MAIL headers  ----------------------- */

     DBGPRINT(DBG$SHOWMORE,"__read_message: getting VMS MAIL headers",0,0,0,0);
     hdrFields = 0;				/* Reset hdr fields flags */
     vmsstarthdr = mbptr;			/* Note start of headers */
     len=__message_readline(stream, mbptr);	/* First record */
     while ( len > 1 )				/* Continue until blank line */
       {
          len = __rewrite_vms_headers(mbptr,len-1);
          if (strncmp(mbptr,"Subject: ",len-1) == 0) 
            hdrFields = hdrFields|fSUBJECT;
          len = __message_readline(stream, mbptr+=len);
       }

						/* Add delivery date field */
     sprintf(mbptr,"Date: %s\n",deliverydate); *(mbptr+8)=' '; *(mbptr+12)=' ';
     mbptr += strlen(mbptr);
     *mbptr++ = '\n';				/* terminate headers */
     vmsheadsize = mbptr-vmsstarthdr;

						/* set date into elt */
     mail_parse_date(elt, deliverydate);

						/* reset rfcheader pointers */
     rfcstarthdr = NULL; rfcheadsize = 0;

  /* -----------------   Possible RFC822  headers  ----------------------- */

     len = __message_readline(stream, mbptr);
     if ( !headstr && (len > 0) && __is_rfc822_header(mbptr,len-1) )
       {
         DBGPRINT(DBG$SHOWMORE,"__read_message: checking for RFC822 headers",0,0,0,0);
         hdrFields = __set_headerFields(mbptr,len-1,hdrFields);
         rfcstarthdr = mbptr;			/* Note start of rfc headers */
         while (len > 1)			/* Continue until blank line */
           {
             len = __message_readline(stream, mbptr+=len);
             hdrFields = __set_headerFields(mbptr,len-1,hdrFields);
           }

						/* it was an X.400 P2 header */
         if ((hdrFields&0x0040) != 0)
           {					/* append VMSmail header */
             memcpy(mbptr,vmsstarthdr,vmsheadsize);
             mbptr += vmsheadsize-1;
             hdrFields = ((hdrFields|fFROM)|fTO)|fSUBJECT;
           }
         *mbptr++ = '\n';			/* terminate headers */
         rfcheadsize = mbptr-rfcstarthdr;

						/* First body line */
         len = __message_readline(stream, mbptr);

						/* additional RFC822 fields */
         if (((hdrFields&0x0040) != 0) &&
             (strncmp(mbptr,"RFC-822-Headers:",len-1) == 0))
           {
             mbptr -= 1;
             while (len > 1)			/* Continue until blank line */
               {
                 len = __message_readline(stream, mbptr);
                 if (*mbptr != 12) mbptr += len;
               }
             rfcheadsize = mbptr-rfcstarthdr;

						/* First true body line */
             len = __message_readline(stream, mbptr);
           }
       }

  /* -----------------   Message body  ----------------------- */

     DBGPRINT(DBG$SHOWMORE,"__read_message: reading message body",0,0,0,0);

						/* Read body until end or header line */
     starttext = mbptr;				/* Remember where body starts */
     while ( (len>0) && !(headstr && ((len-1) == headstrlen) && (strncmp(mbptr,headstr,headstrlen) == 0)) )
         len = __message_readline(stream,mbptr+=len);
     bodysize=mbptr-starttext;
	

  /* -----------------   Possible RFC822 headers at the end ----------- */

						/* Header expected now ? */
     if (headstr)
       {
          DBGPRINT(DBG$SHOWMORE,"__read_message: looking for 'at end' string '%s'", headstr,0,0,0);
          rfcstarthdr = mbptr;
          if (len > 0 )
            {
              DBGPRINT(DBG$SHOWMORE,"__read_message: reading RFC822 headers at end",0,0,0,0);
              len = __message_readline(stream,mbptr);
              hdrFields = __set_headerFields(mbptr,len-1,hdrFields);
              while (len > 0)			/* Continue until EOM */
                {
                  len = __message_readline(stream, mbptr+=len);
                  hdrFields = __set_headerFields(mbptr,len-1,hdrFields);
                }
              *mbptr++ = '\n';
            }

						/* If none found, use original VMS */
          rfcheadsize = mbptr - rfcstarthdr;

          if (rfcheadsize == 0)
            DBGPRINT(DBG$SHOWMORE,"__read_message: no 'at end' headers, using VMS headers", 0,0,0,0);
        }

  /* -----------------  Choose which headers to use -------------------- */

    if ((rfcstarthdr != NULL) && (rfcheadsize > 0))
      if (((hdrFields&0x0003) != 0) && ((hdrFields&0x001C) != 0) &&
          ((hdrFields&0x0020) != 0))
        {
          DBGPRINT(DBG$SHOWMORE,"__read_message: will use RFC822 headers",0,0,0,0);
          starthdr = rfcstarthdr; headsize = rfcheadsize;
        }
      else
        {
          DBGPRINT(DBG$SHOWMORE,"__read_message: will use VMSMail headers",0,0,0,0);
          starthdr = vmsstarthdr; headsize = vmsheadsize;
          if (!headstr)		/* Include false header as text */
            { starttext = rfcstarthdr; bodysize=bodysize+rfcheadsize; }
        }
    else
      {
        DBGPRINT(DBG$SHOWMORE,"__read_message: will use VMSMail headers",0,0,0,0);
        starthdr = vmsstarthdr; headsize = vmsheadsize;
      }

  /* -----------------  Save in cache ---------------------------------- */

						/* Save headers in cache */
     DBGPRINT(DBG$SHOWMORE,"__read_message: caching headers, size='%d'",headsize,0,0,0);
     fc->headersize = headsize;			/* How big ?? */
     fc->header     = (char *) fs_get(headsize+1);
     memmove(fc->header, starthdr, headsize);

						/* Save body in cache */
     DBGPRINT(DBG$SHOWMORE,"__read_message: caching body, size='%d'",bodysize,0,0,0);
     fc->size  = bodysize;			/* How big ?? */
     fc->text  = (char *) fs_get(bodysize + 1);
     memmove(fc->text, starttext, bodysize);

     DBGPRINT(DBG$EXIT,"__read_message, exit, message size='%ld'",headsize+bodysize,0,0,0);
  }

/* VMS/MAIL mail lock file for parse/append permission
 * Accepts: file descriptor
 *	    lock file name buffer
 *	    type of locking operation (LOCK_SH or LOCK_EX)
 * Returns: file descriptor of lock or -1 if failure
 */

static int vms_mail_lock (FILE *fd, char *lock, int op)
{
 return 0;		/* no locking needed */
}


/* VMS/MAIL mail unlock file for parse/append permission
 * Accepts: file descriptor
 *	    lock file name from vms_mail_lock()
 */

static void vms_mail_unlock (FILE *fd, char *lock)
{
  return;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Add a newly created mailfile/folder combination to the list of 'new'
	mailboxes.

   Inputs:
	mailfile		Pointer to name of mailfile
	folder			Pointer to name of folder in mailfile

   Outputs:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void __Add_New_Mailbox(char *mailfile, char *folder)
  {
     struct MailBoxList *mb = fs_get( sizeof(struct MailBoxList) );

     DBGPRINT(DBG$ENTRY,"__Add_New_Mailbox, entered, mailfile='%s',folder='%s'",mailfile,folder,0,0);


					/* Link to head of chain */
     mb->next = MailBoxes;

					/* Copy items into structure */
     mb->mailfile = fs_get( strlen(mailfile) + 1);
     strcpy(mb->mailfile, mailfile);
     mb->folder   = fs_get( strlen(folder)   + 1);
     strcpy(mb->folder, folder);

					/* Link master pointer */
     MailBoxes = mb;

     DBGPRINT(DBG$EXIT,"__Add_New_Mailbox: exit",0,0,0,0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Search the list of 'created' mailboxes

   Inputs:
	mailfile		Pointer to name of mail file
	folder			Pointer to name of folder

   Outputs:
	NONE

   Return Value:
	T			if found
	NIL			if not found

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __search_new_mailboxes(char *mailfile, char *folder)
  {
     struct MailBoxList *mb;

     DBGPRINT(DBG$ENTRY,"__search_new_mailboxes: entered, mailfile='%s', folder='%s'",mailfile,folder,0,0);
     for (mb=MailBoxes; mb; mb=mb->next)
       {
         DBGPRINT(DBG$SHOWMORE,"__search_new_mailboxes: checking mailfile='%s', folder='%s'",mb->mailfile,mb->folder,0,0);
         if ( (strcasecmp(mb->mailfile, mailfile)   ) &&
              (strcmp(mb->folder,   folder)   == 0) )
           {
             DBGPRINT(DBG$EXIT,"__search_new_mailboxes: exit, found",0,0,0,0);
             return T;
           }
        }

     DBGPRINT(DBG$EXIT,"__search_new_mailboxes: exit, not found",0,0,0,0);
     return NIL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Process 'new' folders in specified mailfile

   Inputs:
	action		Ptr to action function
	fp		Ptr to folder pattern structure

   Outputs:
	fp		Ptr to folder pattern structure

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, May 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __scan_new_mailboxes(int (*action)(), struct FolderPattern *fp)
  {
     struct MailBoxList *mb;
     struct DESC         FolderName;

     DBGPRINT(DBG$ENTRY,"__scan_new_mailboxes: entered, mailfile='%s'",fp->mailfile,0,0,0);
     for (mb=MailBoxes; mb; mb=mb->next)
       {
         DBGPRINT(DBG$SHOWMORE,"__scan_new_mailboxes: checking mailfile='%s', folder='%s'",mb->mailfile, mb->folder,0,0);
         if ( strcasecmp(mb->mailfile, fp->mailfile) )
           {
             DBGPRINT(DBG$SHOWMORE,"__scan_new_mailboxes: processing folder '%s'", mb->mailfile, 0, 0, 0);
             FolderName.address=mb->folder;
             FolderName.length=strlen(mb->folder);
             (*action)(fp, &FolderName);
           }
        }
     DBGPRINT(DBG$EXIT,"__scan_new_mailboxes: exit",0,0,0,0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Remove a mailbox from the special list of 'created' mailboxes. This
	should be called when a mailbox is deleted.

   Inputs:
	mailfile	The name of the mailfile
	folder		The name of the folder

   Outputs:
	NONE

   Function Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, May 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __Delete_New_Mailbox(char *mailfile, char *folder)
  {
     struct MailBoxList **mbptr = &MailBoxes;

     DBGPRINT(DBG$ENTRY, "__Delete_New_Mailbox: mailfile=%s, folder=%s", mailfile, folder, 0, 0);     

				/* Scan the list */
     while (*mbptr)
       {
				/* Does this mailbox match ? */
         DBGPRINT(DBG$SHOWMORE,"__Delete_New_Mailbox: cur link: ptr=0x%08x, link=0x%08x, next=0x%08x", mbptr, *mbptr, (*mbptr)->next, 0);
         if ( (strcasecmp((*mbptr)->mailfile, mailfile)   ) &&
              (strcmp((*mbptr)->folder,   folder)   == 0) )
            { 
				/* temp copy of CURRENT pointer */
              struct MailBoxList *mb = *mbptr;

				/* Relink next to previous, skip current */
              *mbptr = (*mbptr)->next;
              DBGPRINT(DBG$SHOWMORE,"__Delete_New_Mailbox: new link: ptr=0x%08x, link=0x%08x", mbptr, *mbptr, 0, 0);

				/* Free up space occupied by current */
              free (mb->mailfile);
              free (mb->folder  );
              free (mb);
	      return;
            }
         mbptr = &(*mbptr)->next;
       }
    DBGPRINT(DBG$EXIT, "__Delete_New_Mailbox: exit, no matching link",0,0,0,0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Initialize the mail stream's local text cache. This cache remembers the
	headers and message body, in text form, for each message as it's read.

   Input:
	stream		The mail stream for which the cache is to be created

   Output:
	NONE

   Return Value:
	NONE

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __init_local_cache(MAILSTREAM *stream)
  {
    long i;
					/* Allocate the message cache */
    LOCAL->msgs = NULL;
    if (stream->nmsgs == 0) return;
    LOCAL->msgs = fs_get(stream->nmsgs * sizeof(FILECACHE));

					/* Clear all entries to null pointers */
    for(i = 0; i < stream->nmsgs; i++) {
	LOCAL->msgs[i].flags      = 0;
        LOCAL->msgs[i].size       = 0;
	LOCAL->msgs[i].text       = NULL;
        LOCAL->msgs[i].headersize = 0;
        LOCAL->msgs[i].header     = NULL;
    }
  }



/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Remove the mail stream's local text cache.

   Input:
	stream		The mail stream for which the cache is to be removed

   Output:
	NONE

   Return Value:
	NONE

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __clear_local_cache(MAILSTREAM *stream)
  {
    long i;
				/* Free all occupied entries */
    for(i = 0; i < stream->nmsgs; i++)
      {
	if(LOCAL->msgs[i].text != NULL)
		fs_give((void **) &LOCAL->msgs[i].text);
	if(LOCAL->msgs[i].header != NULL)
		fs_give((void **) &LOCAL->msgs[i].header);
      }
				/* Free the cache table itself */
    fs_give((void **) &LOCAL->msgs);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Action routine to process each folder found by the FIND ALL command
	It simply notifies the application of the folder name, via the
	mm_mailbox callback.

   Inputs:
	fpi			Folder pattern information
	FolderName		Name of latest folder found in mailbox

   Outputs:
	fpi			As above; contains return status flag

   Return Value:
	Ignored

   Side Effects:
	Calls mm_mailbox with the foldername; the application may do something
	with this name.

   Comments:
	The fpi structure contains the pattern to be matched against and
	returns the count of matches found. Any folder which matches the
	pattern is notified to the application using the mm_mailbox callback.
	The matched name is prefixed by the string in the data field of the fpi
	structure

	We hide the real name of the INBOX folder.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __match_folders(struct FolderPattern *fpi, struct DESC *FolderName)
  {
    char mailbox[MAILFILESIZE];
    int inbox;

    if ( (FolderName==NULL) || (FolderName->length == 0) )/* End of list */
      {
        DBGPRINT(DBG$EXIT,"__match_folders: end of folder list",0,0,0,0);
        return 0;
      }

					/* Get the file mailbox name */
    sprintf(mailbox, "%s%.*s", fpi->prefix, FolderName->length, FolderName->address);
    
					/* Register if it matches the pattern */
    if (matchwild(fpi->pattern, mailbox) == T)
      {
					/* Unless it is what INBOX maps to */
        if (__parse_mailbox_name(mailbox, NULL, NULL, &inbox) && !inbox)
          {
            fpi->count++;
            mm_mailbox( mailbox );
          }
      }

    DBGPRINT(DBG$EXIT,"__match_folders: pattern='%s', count='%3d', folder='%.*s'", fpi->pattern, fpi->count, FolderName->length, FolderName->address);
    return 1;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Action routine to locate a particular folder within a mail file.

   Inputs:
	fpi			Folder pattern information
	FolderName		Name of latest folder found in mailbox

   Outputs:
	fpi			As above; contains return status flag

   Return Value:
	Ignored

   Comments:
	The fpi structure contains the folder name required. If found, the
	count field is incremented.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __find_folder(struct FolderPattern *fpi, struct DESC *nextfolder)
  {
				 	/* End of list */
    if( (nextfolder==NULL)  ||  (nextfolder->length == 0) )
      {
        DBGPRINT(DBG$EXIT,"__find_folder: end of folder list",0,0,0,0);
        return 0;
      }

					/* See if this matches what we want */
    if (nextfolder->length == strlen(fpi->pattern))
      if ( strncmp(nextfolder->address, fpi->pattern, nextfolder->length) == 0)
        fpi->count++;


    DBGPRINT(DBG$EXIT,"__find_folder: '%s': count '%3d', folder '%.*s'", fpi->pattern, fpi->count, nextfolder->length, nextfolder->address);
    return 1;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Scan the folders in a mailfile and call an action routine to deal
	with each one. This is really a simple jacket routine for the
	mail$mailfile_info_file routine.

   Inputs:
	context		Address of the mailfile context
	routine		Address of function to call
	userdata	Address of userdata to pass

   Outputs:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void __scan_folders(int *context, int (*action)(), void *userdata)
  {
    int status;

    item ScanFolders[] = {
	{sizeof(action),   MAIL$_MAILFILE_FOLDER_ROUTINE, action,    NULL},
	{sizeof(userdata), MAIL$_MAILFILE_USER_DATA,      userdata, NULL},
        {0,                0,                             NULL,      NULL}};

    DBGPRINT(DBG$ENTRY, "__scan_folders: entered, action = '0x%08x', userdata='0x%08x'",action, userdata, 0, 0);
    DBGPRINT(DBG$SHOWMORE,"__scan_folders: scanning 'new' list",0,0,0,0);
    __scan_new_mailboxes(action, userdata);
    DBGPRINT(DBG$SHOWMORE,"__scan_folders: scanning folders in mailfile",0,0,0,0);
    status = mail$mailfile_info_file(context, ScanFolders, NullItemList);
    chkstat("mail$mailfile_info_file", 1, status);
    DBGPRINT(DBG$EXIT, "__scan_folders: exit",0,0,0,0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Copy a message from the current message/folder to a new mailbox

   Inputs:
	stream		The mail stream for the message source
	msgno		The message number
	outmfile	Ptr to name of destination mailfile
	outfolder	Ptr to name of destination folder

   Outputs:
	NONE

   Return Value:
	T		If copy OK
	NIL		If copy not OK

   Author:
	Andy Harper, Kings College London, November 1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int vms_mail_message_copy(MAILSTREAM *stream, long msgno, char *outmfile, char *outfolder)
  {
    static int FileCreated, FoldCreated;
    int status;

    item CopyFolderItem[] = {
	{strlen(outfolder),   MAIL$_MESSAGE_FOLDER,         outfolder,    NULL},	/* must be 1st,foldername */
	{strlen(outmfile),    MAIL$_MESSAGE_FILENAME,       outmfile,     NULL},	/* must be 2nd */
	{sizeof(msgno),       MAIL$_MESSAGE_ID,             &msgno,       NULL},	/* Message number to move */
	{0, 0, NULL, NULL } };
    
    item CreatedItemList[] = {
	{sizeof(FileCreated), MAIL$_MESSAGE_FILE_CREATED,   &FileCreated, NULL},
	{sizeof(FoldCreated), MAIL$_MESSAGE_FOLDER_CREATED, &FoldCreated, NULL},
	{0, 0, NULL, NULL } };

    DBGPRINT(DBG$ENTRY,"vms_mail_message_copy: entered, source='%s', msgno='%ld', destination='%s -- %s'", stream->mailbox, msgno, outmfile, outfolder);
    
    status = mail$message_copy(&LOCAL->msg_context, CopyFolderItem, CreatedItemList);
    if (!chkstat("mail$message_copy", 0, status) ) return NIL;
    if (FileCreated) mm_log("Created mailbox file", WARN);
    if (FoldCreated) mm_log("Created mailbox folder", WARN);

    DBGPRINT(DBG$EXIT,"vms_mail_message_copy: exit",0,0,0,0);
    return T;
  }

/* VMS/MAIL copy messages
 * Accepts: MAIL stream
 *	    mailbox copy vector
 *	    mailbox name
 * Returns: T if success, NIL if failed
 */

long vms_mail_copy_messages (MAILSTREAM *stream, char *mailbox)
  {
    char outmfile[MAILFILESIZE],outfolder[FOLDERSIZE];
    int inbox;
    long i;

    DBGPRINT(DBG$ENTRY,"vms_mail_copy_messages: entered, stream='%s', destination='%s'", stream->mailbox, mailbox, 0, 0);

    if (__parse_mailbox_name(mailbox, outmfile, outfolder, &inbox) == NIL)
      {
        mm_log("invalid destination folder", ERROR);
        return NIL;
      }

    for (i=1; i <= stream->nmsgs; i++)
      if (mail_elt(stream,i)->sequence)
        if (vms_mail_message_copy(stream,i,outmfile,outfolder) == NIL)
          {
            mm_log("message copy failed", ERROR);
            return NIL;
          }

    DBGPRINT(DBG$EXIT,"vms_mail_copy_messages: exit",0,0,0,0);
    return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	If this is the INBOX folder and the auto-file inbox flag is set, then
	move all SEEN messages from the INBOX to the specified mailbox.

   Input:
	stream		The mail stream to be checked

   Output:
	NONE

   Return Value:
	NONE

   Comments:
	This is normally called during a close operation and will normally have
	been preceded by an expunge operation, so some messages flagged as SEEN
	will already have been deleted!

	The LOCAL->inbox flag is used to check if this an inbox

	The logical name CCLIENT_AUTOFILEINBOX defines which mailbox the messages
	are to be moved into. If it is not defined, then INBOX messages are not
	moved (leave undefined for standard PINE/IMAPD compatibility, define to
	"MAIL" for VMS MAIL compatibility).

    Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __file_inbox_messages(MAILSTREAM *stream)
  {
    char *nf, mailfile[MAILFILESIZE], folder[FOLDERSIZE];
    int inbox,i,f;
    long filed=0;

    if (!stream) return;
    if (!LOCAL)  return;

    DBGPRINT(DBG$ENTRY,"__file_inbox_messages: entered, mailbox='%s'", stream->mailbox, 0, 0, 0);


				/* Is this the INBOX ? */
    if (!LOCAL->inbox)
      {
        DBGPRINT(DBG$EXIT,"__file_inbox_messages: exit, folder not INBOX", 0, 0, 0, 0);
        return;
      }

				/* Is the autofile option set up? */
    if ((nf = getenv("CCLIENT_AUTOFILEINBOX")) == NULL)
      {
        DBGPRINT(DBG$EXIT,"__file_inbox_messages: exit, auto-file logical not defined", 0, 0, 0, 0);
        return;
      }

				/* Validate mailbox specification */
    if ( __parse_mailbox_name(nf, mailfile, folder, &inbox) == NIL)
      {
        DBGPRINT(DBG$EXIT,"__file_inbox_messages: exit, invalid destination mailbox",0,0,0,0);
        return;
      }


				/* Delete messages SEEN but not DELETED */
    DBGPRINT(DBG$SHOWMORE,"__file_inbox_messages: filing seen messages to '%s'", nf, 0, 0, 0);
    for(i = 1; i <= stream->nmsgs; i++)
     {
       f = LOCAL->msgs[i-1].flags;
       if( (f & fSEEN)  &&  !(f & fDELETED) )
        {
					/* Copy the message */
          DBGPRINT(DBG$SHOWMORE,"__file_inbox_messages: filing message '%d'", i,0,0,0);
          if (vms_mail_message_copy(stream, i, mailfile, folder) == NIL)
            {
              mm_log("Auto-filing of INBOX messages not completed",WARN);
              return;
            }
					/* Flag it as deleted */
          __setflag(stream, i, mail_elt(stream,i), fDELETED);
          filed++;
        }
     }

					/* Remove messages flagged for delete */
    if (filed)
      { char temp[256];
        sprintf(temp,"auto-filed %d messages to %s", filed, nf);
        mm_log(temp, NIL);
        vms_mail_expunge (stream);
      }    
    DBGPRINT(DBG$EXIT,"__file_inbox_messages: exit",0,0,0,0);
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Open a specific mail file and return the context value

   Input:
	MailFile		The name of the mailfile to open

   Output:
	NONE

   Return Value:
	The callable mail context; or zero if failed

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static int __create_mail_context(char *MailFile)
  {
    int status, context=0;

    item MailFileList[] = {	/* Set the name of the file */
	{ strlen(MailFile), MAIL$_MAILFILE_NAME, MailFile, NULL},
        { 0,                MAIL$_NOSIGNAL,      NULL,     NULL},
	{ 0, 0, NULL, NULL}};

    DBGPRINT(DBG$ENTRY,"__create_mail_context: Entered.. File='%s'", MailFile,0,0,0);


					/* Set up mailfile context */
    status = mail$mailfile_begin(&context, NullItemList, NullItemList);
    if (!chkstat("mail$mailfile_begin", 0, status)) return 0;
    status = mail$mailfile_open(&context, MailFileList, NullItemList);
    if (!chkstat("mail$mailfile_open", 0, status))
      {
        status = mail$mailfile_end(&context, NullItemList, NullItemList);
        chkstat("mail$mailfile_end", 0, status);
        return 0;
      }


    DBGPRINT(DBG$EXIT,"__create_mail_context: exiting, mail ctx=0x%08x", context, 0, 0, 0);
    return context;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Initialize the message context for an opened mail file

   Input:
	mailcontext		Pointer to the mail context for the open file

   Output:
	NONE

   Return Value:
	The message context; or zero if failure

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __create_msg_context(int *mailcontext)
  {
    int status, context=0;

    item MessageContext[] = {
	{sizeof(*mailcontext), MAIL$_MESSAGE_FILE_CTX, mailcontext, NULL},
	{0,                    0,                      NULL,        NULL}};

    DBGPRINT(DBG$ENTRY,"__create_msg_context: entered, mailcontext = '0x%08x'", *mailcontext, 0, 0, 0);
    status = mail$message_begin(&context, MessageContext, NullItemList);
    if (!chkstat("mail$message_begin", 0, status)) return 0;
    DBGPRINT(DBG$EXIT,"__create_msg_context: exit, messagecontext='0x%08x'",context,0,0,0);
    return context;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Close the mailfile by ending the specified message and mail contexts

   Inputs:
	mailcontext		Pointer to current mail file context
	messagecontext		Pointer to current message context

   Outputs:
	mailcontext		As above; set to zero
	messagecontext		As above; set to zero

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __vms_mail_close(int *mailcontext, int *messagecontext)
  {

    DBGPRINT(DBG$ENTRY,"__vms_mail_close: Entered, mail context='0x%08x', messagecontext=0x%08x'",*mailcontext,*messagecontext,0,0);
    if (*mailcontext != 0)
      {
	int status;

        item CloseItem[] = {	/* Purge and reclaim during close */
		{ 0, MAIL$_MAILFILE_FULL_CLOSE, NULL, NULL},
		{ 0, MAIL$_NOSIGNAL,            NULL, NULL},
		{ 0, 0, NULL, NULL } };

        int purgereclaim = getenv("CCLIENT_MAILFILE_PURGE_RECLAIM");

				/* Terminate message context (if any) */
        if (*messagecontext != 0)
          {
            DBGPRINT(DBG$SHOWMORE,"__vms_mail_close: Terminating message context (0x%08x)",*messagecontext,0,0,0);
    	    status = mail$message_end(messagecontext, NullItemList, NullItemList);
            chkstat("mail$message_end", 0, status);
          }

				/* Terminate the mail context */
        DBGPRINT(DBG$SHOWMORE,"__vms_mail_close: Terminating mail context (0x%08x)",*mailcontext,0,0,0);
        status = mail$mailfile_close(mailcontext, purgereclaim ? CloseItem : NullItemList, NullItemList);
        chkstat("mail$mailfile_close", 0, status);
        status = mail$mailfile_end(mailcontext,   purgereclaim ? CloseItem : NullItemList, NullItemList);
        chkstat("mail$mailfile_end", 0, status);
      }
    DBGPRINT(DBG$EXIT,"__vms_mail_close: exiting, mail ctx=0x%08x, msg ctx=0x%08x", *mailcontext, *messagecontext, 0, 0);
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Test whether a particular mailbox exists by opening the mailfile and
	scanning for the folder.

   Inputs:
	mailbox			The mailbox to check
	mailcontext		Pointer to existing mail context, or NULL
	close			Leave open, or close when done

   Output:
	mailfile		Pointer to parsed mailfile buffer; or NULL
	folder			Pointer to parsed folder buffer; or NULL
	inbox			Pointer to 'is the INBOX' flag
		
   Return Value:
	T			mailbox exists
	NIL			mailbox does not exist

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __folder_exists(char *mailbox, int *mailcontext, int close, char *mailfile, char *folder, int *inbox)
  {
    char tmpmailfile[MAILFILESIZE], tmpfolder[FOLDERSIZE];
    struct stat sbuf;
    struct FolderPattern fp = {0, tmpmailfile, NULL, tmpfolder};

    int dummycontext = 0;		/* For closing */
    int context      = 0;		/* internal context if none supplied */
    int *ctx         = &context;	/* Context pointer */

    DBGPRINT(DBG$ENTRY,"__folder_exists: entered, mailbox='%s'",mailbox,0,0,0);

					/* get mailfile and foldername */
    if (!__parse_mailbox_name(mailbox, tmpmailfile, tmpfolder, inbox))
      {
        DBGPRINT(DBG$EXIT,"__folder_exists: exit, cannot parse",0,0,0,0);
        return NIL;
      }


    if (stat(tmpmailfile, &sbuf) == -1)
      {
        DBGPRINT(DBG$EXIT,"__folder_exists: exit, cannot stat mailfile",0,0,0,0);
        return NIL;
      }

					/* Copy names, if wanted */
    if (mailfile)
      strcpy(mailfile,tmpmailfile);
    if (folder)
      strcpy(folder,tmpfolder);

					/* Use external context if given */
    if (mailcontext)
      ctx = mailcontext;

					/* If not open, open it */
    if (!*ctx)
      {
        DBGPRINT(DBG$SHOWMORE,"__folder_exists: opening '%s'", tmpmailfile,0,0,0);
        if ((*ctx = __create_mail_context(tmpmailfile)) == 0)
          {
             DBGPRINT(DBG$EXIT,"__folder_exists: exit, open failed",0,0,0,0);
             return NIL;
          }
      }

					/* Scan the folder list */
    DBGPRINT(DBG$SHOWMORE,"__folder_exists: scanning '%s' for folder '%s'",tmpmailfile,tmpfolder,0,0);
    __scan_folders(ctx, __find_folder, &fp);


					/* Close if internal or close requested */
    if (context || close)
      {
        DBGPRINT(DBG$SHOWMORE,"__folder_exists: closing mail file",0,0,0,0);
        __vms_mail_close(ctx, &dummycontext);
      }

					/* Return T if found, else NIL */
    DBGPRINT(DBG$EXIT,"__folder_exists: exit, count ='%d'",fp.count,0,0,0);
    return (fp.count>0) ? T : NIL;
  }

/* VMS/MAIL manipulate driver parameters
 * Accepts: function code
 *	    function-dependent value
 * Returns: function-dependent return value
 */

void *vms_mail_parameters (long function, void *value)
{
  return NULL;
}



/* VMS/MAIL mail validate mailbox
 * Accepts: mailbox name
 * Returns: our driver if name is valid, NIL otherwise
 */

DRIVER *vms_mail_valid (char *name)
{
  return vms_mail_isvalid(name) ? &vmsmaildriver : NIL;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Validate the mailbox. Check that the syntax is right for this driver
	and that the mailbox exists.

   Input:
	name		Pointer to mailbox name

   Output:
	NONE

   Return Value:
	T if mailbox exists
	NIL if it doesn't or the format is invalid

  Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
long vms_mail_isvalid (char *name)
{
  static char *invalchrs = "{}()[]^&";
  char *i;
  int inbox;

  DBGPRINT(DBG$ENTRY,"vms_mail_isvalid: entered, name='%s'", name,0,0,0);

						/* Invalid chars ?? */
  for (i=invalchrs; *i; i++)
    if (strchr(name,*i))
      {
        DBGPRINT(DBG$EXIT,"vms_mail_isvalid: mailbox name has invalid char '%c'",*i,0,0,0);
        return NIL;
      }

						/* Successful parse ?? */
  if (__folder_exists(name, NULL, 1, NULL, NULL, &inbox))
    {
      DBGPRINT(DBG$EXIT,"vms_mail_isvalid: exiting, mailbox name OK",0,0,0,0);
      return T;
    }
  else
    {
      DBGPRINT(DBG$EXIT,"vms_mail_isvalid: could not find mailbox '%s'",name,0,0,0);
      return NIL;
    }

}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Display all subscribed mailboxes, from the subscription database, which
	match the supplied pattern.

   Inputs:
	stream		The mail stream (not used)
	pat		The pattern to match

   Outputs:
	NONE

   Return Value:
	NONE

   Comments:
	results are reported back using the mm_mailbox() callback.

	The list of subscribed mail folders is maintained in a database file
	in the user's home directory (see define of MBXDATABASE)

	If the INBOX exists, it is always returned

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
void vms_mail_find (MAILSTREAM *stream, char *pat)
{
  char *home, subscriptionfile[MAILFILESIZE], record[MAXRECORDSIZE];
  FILE *fp;

					/* Always have an INBOX if it exists */
  if (matchwild(pat,"INBOX") && vms_mail_isvalid("INBOX"))
    mm_mailbox("INBOX");

					/* Find location of subscription database */
  if (__get_home_directory(&home) == NULL)
    {
      mm_log("cannot locate subscription database",ERROR);
      return;
    }

					/* Scan subscription database */
  strcpy(subscriptionfile,home);
  strcat(subscriptionfile,MBXDATABASE);
  if ((fp=fopen(subscriptionfile,"r")) != NULL)
    {
      while ( fgets(record, MAXRECORDSIZE, fp)  )
       {
        record[strlen(record)-1] = '\0';/* strip '\n' */
        if ( matchwild(pat, record) )
           mm_mailbox(record);
       }
      fclose(fp);
    }

  return;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Find all the mailboxes that match the pattern. This actually returns
	all folders in the mail file(s) that match the pattern.

   Inputs:
	stream		The mail stream (not used)
	pat		The pattern to match

   Outputs:
	NONE

   Return Value:
	NONE

   Comments:
	Results are reported back via the mm_mailbox() callback.

	The nominated mail file is scanned and all matching folders are
	reported.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
void vms_mail_find_all (MAILSTREAM *stream, char *pat)
{
  char *p, prefix[MAILFILESIZE], mailfile[MAILFILESIZE];
  int context=0, dummymessagecontext=0, inbox;
  struct FolderPattern fp = {0, mailfile, "", pat};

  DBGPRINT(DBG$ENTRY,"vms_mail_find_all: entered, pattern = '%s'", pat,0,0,0);


					/* Find possible mailfile name prefix */
   strcpy(prefix, pat);
   if ((p = strrchr(prefix,'/')) != NULL)
     {
       *(p+1) = '\0';
       DBGPRINT(DBG$SHOWMORE,"vms_mail_find_all: found mailbox prefix '%s'", prefix, 0, 0, 0);
       fp.prefix = prefix;
     }

					/* parse mailbox pattern */
  if (__parse_mailbox_name(pat, mailfile, NULL, &inbox) == NIL)
    {
       mm_log("Invalid folder pattern", ERROR);
       DBGPRINT(DBG$EXIT,"vms_mail_find_all: exit : invalid pattern",0,0,0,0);
       return;
    }

					/* Open the mail file */
  if ( (context=__create_mail_context(mailfile)) == 0 )
    {
      DBGPRINT(DBG$EXIT,"vms_mail_find_all: exit : cannot open mail file '%s'", mailfile,0,0,0);
      return;
    }

					/* Find matching folders */
  __scan_folders(&context, __match_folders, &fp);
  __vms_mail_close(&context,&dummymessagecontext);	/* Close it */

  DBGPRINT(DBG$EXIT,"vms_mail_find_all: exit",0,0,0,0);
}

/* VMS/MAIL mail find list of bboards
 * Accepts: mail stream
 *	    pattern to search
 */

void vms_mail_find_bboards (MAILSTREAM *stream, char *pat)
{
  return;		/* no bboards supported on VMS */
}



/* VMS/MAIL mail find list of all bboards
 * Accepts: mail stream
 *	    pattern to search
 */

void vms_mail_find_all_bboards (MAILSTREAM *stream,char *pat)
{
   return;		/* Again, not supported at this stage */
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Add a new mailbox to the list of subscribed folders in the subscription
	database.

   Inputs:
	stream		The mail stream (not used)
	mailbox		Name of mailbox to add

   Outputs:
	NONE

   Return Value:
	T		If subscription accepted
	NIL		If subscription rejected

   Comments:
	Since INBOX is automatically included on the search of the subscription
	database, it may not be added to the file.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

long vms_mail_subscribe (MAILSTREAM *stream, char *mailbox)
{
  char *home,subscriptionlist[MAILFILESIZE], record[MAXRECORDSIZE];
  FILE *fp;

					/* INBOX has special treatment */
  if (strcasecmp(mailbox, "INBOX"))
    {
      mm_log("INBOX already implictly subscribed",ERROR);
      return NIL;
    }

					/* Get home directory */
  if (__get_home_directory(&home) == NULL)
    {
       mm_log("can't locate subscription database",ERROR);
       return NIL;
    }

					/* Construct subscription database name */
  strcpy(subscriptionlist, home);
  strcat(subscriptionlist,MBXDATABASE);


					/* Do we have a subscription database ?? */
  if ((fp=fopen(subscriptionlist,"r")) != NULL)
    {
					/* Is mailbox already subscribed? */
      while ( fgets(record, MAXRECORDSIZE, fp)  )
        {
          record[strlen(record)-1] = '\0';	/* remove '\n' */
          if ( strcmp(mailbox,record) )
            {
              fclose(fp);
              sprintf(record,"already subscribed to %s",mailbox);
              mm_log(record,ERROR);
              return NIL;
            }
        };
      fclose(fp);
    }


						/* Append new subscription */
   if ((fp=fopen(subscriptionlist,"a","rfm=var","rat=cr")) == NULL)
     {
       mm_log("unable to append to subscription database",ERROR);
       return NIL;
     }
   fputs(mailbox,fp); fputc('\n',fp);
   fclose(fp);
   return T;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Delete a mailbox from the mailbox subscription database

   Input:
	stream		Mail stream (not used)
	mailbox		Mailbox to unsubscribe

   Output:
	NONE

   Return Value:
	T		Subscription successfully removed
	NIL		Subscription not removed

   Comments:
	The name is removed from the subscription file in the user's home
        directory.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

long vms_mail_unsubscribe (MAILSTREAM *stream, char *mailbox)
{
  char subscriptionfile[MAILFILESIZE], newsubscriptionfile[MAILFILESIZE], oldsubscriptionfile[MAILFILESIZE];
  char *home, record[MAXRECORDSIZE];
  FILE *fp,*newfp;
  int subscribed;

					/* INBOX not valid */
  if ( strcasecmp("INBOX", mailbox) )
    {
      mm_log("cannot unsubscribe INBOX",ERROR);
      return NIL;
    }

					/* Get location of subscription database */
  if (__get_home_directory(&home) == NULL)
    {
      mm_log("cannot locate subscription database",ERROR);
      return NIL;
    }

					/* Construct subscription database names */
  strcpy(subscriptionfile, home);
  strcat(subscriptionfile, MBXDATABASE);
  
  strcpy(oldsubscriptionfile, subscriptionfile);
  strcat(oldsubscriptionfile, "-OLD");
  strcpy(newsubscriptionfile, subscriptionfile);
  strcat(newsubscriptionfile, "-NEW");


					/* Open existing subscriptions */
  if ((fp=fopen(subscriptionfile,"r")) == NULL)
    {
      mm_log("no subscription file", ERROR);
      return NIL;
    }


					/* Create new subscription database */
  if ((newfp=fopen(newsubscriptionfile,"w","rfm=var","rat=cr")) == NULL)
    {
      fclose(fp);
      mm_log("cannot create new subscription database",ERROR);
      return NIL;
    }
      
					/* Copy, excluding one to delete */
  subscribed = 0;
  while ( fgets(record, MAXRECORDSIZE, fp) )
    {
       record[strlen(record)-1] = '\0';

       if (strcmp(mailbox,record))
	  { subscribed++; }
       else
          { fputs(record,newfp); fputc('\n',newfp);}
    }
  fclose(fp);
  fclose(newfp);


					/* Did we find the mailbox ?? */
  if (!subscribed)
    {
      remove(newsubscriptionfile);
      mm_log("mailbox not subscribed",ERROR);
      return NIL;
    }


					/* Preserve old subscriptions */
  if (rename(subscriptionfile, oldsubscriptionfile) != 0)
    {
      remove(newsubscriptionfile);
      mm_log("cannot save old subscription database",ERROR);
      return NIL;
    }


					/* Rename new copy to master */
  if (rename(newsubscriptionfile, subscriptionfile) != 0)
    {
       remove(newsubscriptionfile);
       rename(oldsubscriptionfile,subscriptionfile);
       mm_log("cannot rename new subscription database",ERROR);
       return NIL;
    }

					/* Get rid of old copy */
  if (remove(oldsubscriptionfile) !=0 )
     mm_log("could not remove old subscription file", WARN);

  return T;
}

/* VMS/MAIL mail subscribe to bboard
 * Accepts: mail stream
 *	    bboard to add to subscription list
 * Returns: T on success, NIL on failure
 */

long vms_mail_subscribe_bboard (MAILSTREAM *stream, char *mailbox)
{
  mm_log("not implemented on VMS",ERROR);
  return NIL;			/* never valid for VMS/MAIL */
}


/* VMS/MAIL mail unsubscribe to bboard
 * Accepts: mail stream
 *	    bboard to delete from subscription list
 * Returns: T on success, NIL on failure
 */

long vms_mail_unsubscribe_bboard (MAILSTREAM *stream, char *mailbox)
{
  mm_log("not implemented on VMS",ERROR);
  return NIL;			/* never valid for VMS/MAIL */
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Create a new mailbox

   Inputs:
	stream		Name of mail stream (not used)
	mailbox		Name of mailbox to create

   Output:
	NONE

   Return Value:
	T		mailbox successfully created
	NIL		mailbox not created

   Comments:
	This routine creates an empty file based on the FDL specification of
	another file:
		Either: the user's default MAIL file
		Or:	The prototype specified by CCLIENT_MAILPROTYPE

	We cannot create a 'folder' in a VMS MAIL file so we only create the
	mail file itself. VMS MAIL does not allow empty folders to exist as
	they are created automatically when a message is copied into them.

	We need to handle the checking carefully. We must check that the named
	folder exists within the mail file.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

long vms_mail_create (MAILSTREAM *stream, char *mailbox)
{
   char mailfile[MAILFILESIZE], folder[FOLDERSIZE], fdlbuffer[4096];
   char tmp[256], maildefault[MAILFILESIZE];
   char *p;
   int fdlsize,i,inbox;
   struct stat sbuf;

   char *prototypes[] = {maildefault,"CCLIENT_MAILPROTOTYPE"};
   int maxproto       = sizeof(prototypes)/sizeof(prototypes[0]);
   int first          = 0;
   int context        = 0;
   int dummycontext   = 0;


   DBGPRINT(DBG$ENTRY,"vms_mail_create: entered, mailbox='%s'", mailbox,0,0,0);

   if (!__parse_mailbox_name("INBOX", maildefault, NULL, &inbox))
     {
       DBGPRINT(DBG$SHOWMORE,"vms_mail_create: can't parse INBOX",0,0,0,0);
       first++;	/* skip attempt to get prototype from INBOX */
     }
      
/*
   DBGPRINT(DBG$SHOWMORE,"vms_mail_create: checking existence of folder",0,0,0,0);
   if (__folder_exists(mailbox, &context, 1, NULL, NULL, &inbox))
     {
       mm_log("folder already exists",ERROR);
       return NIL;
     }
*/

   DBGPRINT(DBG$SHOWMORE,"vms_mail_create: parsing mailbox",0,0,0,0)
   if (!__parse_mailbox_name(mailbox,mailfile,folder,&inbox))
     {
       mm_log("unparseable mailbox name",ERROR);
       return NIL;
     }

   DBGPRINT(DBG$SHOWMORE,"vms_mail_create: checking existence of mail file",0,0,0,0);
   if (stat(mailfile, &sbuf) == 0)
     {
       __Add_New_Mailbox(mailfile,folder);
       DBGPRINT(DBG$SHOWMORE,"vms_mail_create: mailfile exists, but no folder",0,0,0,0);
       return T;
     }
  
   
   DBGPRINT(DBG$SHOWMORE,"vms_mail_create: creating directory", 0, 0, 0, 0);
   if ((p=strchr(mailfile,']')) != NULL)
     {
       int len = (p-mailfile) + 1;
       DBGPRINT(DBG$SHOWMORE,"vms_mail_create: dirlen = '%d'", len,0,0,0);
       strncpy(tmp, mailfile, len);
       tmp[len] = '\0';
       DBGPRINT(DBG$SHOWMORE,"vms_mail_create: directory is '%s'", tmp,0,0,0);
       if (mkdir(tmp,0) != 0)
         {
           DBGPRINT(DBG$SHOWMORE,"vms_mail_create: directory create failed",0,0,0,0);
         }
     }

   DBGPRINT(DBG$SHOWMORE,"vms_mail_create: attempting create from prototype",0,0,0,0);
   for (i=first; i<maxproto; i++)
     {
       if (!__fdl_generate(prototypes[i], fdlbuffer, sizeof(fdlbuffer)-1, &fdlsize)) continue;
       if (!__fdl_create_file(mailfile, fdlbuffer, fdlsize)) continue;
       __Add_New_Mailbox(mailfile,folder);
       DBGPRINT(DBG$EXIT,"vms_mail_create: exit, created mailfile",0,0,0,0);
       return T;
     }

   sprintf(tmp,"could not create '%s'",mailbox);
   mm_log(tmp, ERROR);
   DBGPRINT(DBG$EXIT,"vms_mail_create: exit, create failed",0,0,0,0);
   return NIL;
}

/* VMS/MAIL mail delete mailbox
 * Accepts: MAIL stream
 *	    mailbox name to delete
 * Returns: T on success, NIL on failure
 */

long vms_mail_delete (MAILSTREAM *stream, char *mailbox)
{
   char mailfile[MAILFILESIZE], folder[FOLDERSIZE];
   int status, msgcnt, n;
   int context = 0, msgctx = 0;

   item OpenFileList[] =
      {{0,               MAIL$_MAILFILE_NAME,    mailfile, NULL},
       {0, 0, 0, 0}};

   item OpenMsgList[] =
      {{sizeof(context), MAIL$_MESSAGE_FILE_CTX, &context,    NULL},
       {0, 0, 0, 0}};

   item SlctMsgList[] =
      {{0,               MAIL$_MESSAGE_FOLDER,   folder,   NULL},
       {0, 0, 0, 0}};

   item SlctCntList[] =
      {{sizeof(msgcnt),  MAIL$_MESSAGE_SELECTED, &msgcnt,     NULL},
       {0, 0, 0, 0}};

   item DeleteItemList[] =
      {{sizeof(n),       MAIL$_MESSAGE_ID,       &n,          NULL},
       {0, 0, 0, 0}};

   DBGPRINT(DBG$ENTRY,"vms_mail_delete: entered, mailbox='%s'", mailbox,0,0,0)

   __parse_mailbox_name(mailbox, mailfile, folder, &n);
   DBGPRINT(DBG$SHOWMORE, "vms_mail_delete: mail=%s, folder=%s", mailfile, folder, 0, 0);
   OpenFileList[0].length  = strlen(mailfile);
   SlctMsgList[0].length   = strlen(folder);

   chkstat("mail$mailfile_begin", 0, mail$mailfile_begin(&context, NullItemList, NullItemList));
   chkstat("mail$mailfile_open",  0, mail$mailfile_open (&context, OpenFileList, NullItemList));
   chkstat("mail$message_begin",  0, mail$message_begin (&msgctx,  OpenMsgList, NullItemList));
   chkstat("mail$message_select", 0, mail$message_select(&msgctx,  SlctMsgList, SlctCntList ));

   for (n=1; n<=msgcnt; n++)
   if (!(status=chkstat("mail$message_delete", 0, mail$message_delete(&msgctx,DeleteItemList,NullItemList))))
     { char tmp[256];
       sprintf(tmp,"Can't delete msg %d from %s (status: 0X%08x)", n, mailbox, status, 0);
       DBGPRINT(DBG$SHOWMORE, "vms_mail_delete: msg %d: delete failed, status=0x%08x",n,status,0,0);
       mm_log(tmp,ERROR);
       return NIL;
     }

   chkstat("mail$message_end:",   0, mail$message_end   (&msgctx,  NullItemList, NullItemList));
   chkstat("mail$mailfile_close", 0, mail$mailfile_close(&context, NullItemList, NullItemList));
   chkstat("mail$mailfile_end",   0, mail$mailfile_end  (&context, NullItemList, NullItemList));

   __Delete_New_Mailbox(mailfile,folder);
   DBGPRINT(DBG$EXIT,"vms_mail_delete: exit",0,0,0,0);
  return T;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Rename a mailbox

   Inputs:
	stream		mail stream to use if needed (it's not)
	old		old mailbox name
	new		new mailbox name

   Output:
	NONE

   Return Value:
	T		if rename ok
	NIL		if rename failed

   Comments:
	Because a mailbox is really a folder inside a file, we can't just
	rename it like a file. We move all messages to the new folder instead.

   Author:
	Andy Harper, Kings College London, May 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

long vms_mail_rename (MAILSTREAM *stream, char *old, char *new)
{
   char srcmailfile[MAILFILESIZE], srcfolder[FOLDERSIZE];
   char dstmailfile[MAILFILESIZE], dstfolder[FOLDERSIZE];
   int status, msgcnt, n;
   int context = 0, msgctx = 0;

   item OpenFileList[] =
      {{0,               MAIL$_MAILFILE_NAME,    srcmailfile, NULL},
       {0, 0, 0, 0}};

   item OpenMsgList[] =
      {{sizeof(context), MAIL$_MESSAGE_FILE_CTX, &context,    NULL},
       {0, 0, 0, 0}};

   item SlctMsgList[] =
      {{0,               MAIL$_MESSAGE_FOLDER,   srcfolder,   NULL},
       {0, 0, 0, 0}};

   item SlctCntList[] =
      {{sizeof(msgcnt),  MAIL$_MESSAGE_SELECTED, &msgcnt,     NULL},
       {0, 0, 0, 0}};

   item MoveItemList[] =
      {{sizeof(n),       MAIL$_MESSAGE_ID,       &n,          NULL},
       {0,               MAIL$_MESSAGE_FILENAME, dstmailfile, NULL},
       {0,               MAIL$_MESSAGE_FOLDER,   dstfolder,   NULL},
       {0,               MAIL$_MESSAGE_DELETE,   NULL,        NULL},
       {0, 0, 0, 0}};

   DBGPRINT(DBG$ENTRY,"vms_mail_rename: entered, old='%s', new='%s'", old,new,0,0)

   __parse_mailbox_name(old, srcmailfile, srcfolder, &n);
   __parse_mailbox_name(new, dstmailfile, dstfolder, &n);
   DBGPRINT(DBG$SHOWMORE, "vms_mail_rename: srcmail=%s, srcfolder=%s",srcmailfile, srcfolder, 0, 0);
   DBGPRINT(DBG$SHOWMORE, "vms_mail_rename: dstmail=%s, dstfolder=%s",dstmailfile, dstfolder, 0, 0);
   OpenFileList[0].length  = strlen(srcmailfile);
   SlctMsgList[0].length   = strlen(srcfolder);
   MoveItemList[1].length  = strlen(dstmailfile);
   MoveItemList[2].length  = strlen(dstfolder);

   chkstat("mail$mailfile_begin", 0, mail$mailfile_begin(&context, NullItemList, NullItemList));
   chkstat("mail$mailfile_open",  0, mail$mailfile_open (&context, OpenFileList, NullItemList));
   chkstat("mail$message_begin",  0, mail$message_begin (&msgctx,  OpenMsgList, NullItemList));
   chkstat("mail$message_select", 0, mail$message_select(&msgctx,  SlctMsgList, SlctCntList ));

   for (n=1; n<=msgcnt; n++)
   if (!(status=chkstat("mail$message_copy", 0, mail$message_copy(&msgctx,MoveItemList,NullItemList))))
     { char tmp[256];
       sprintf(tmp,"Can't move msg %d from %s to %s", n, old, new, status);
       DBGPRINT(DBG$SHOWMORE, "vms_mail_rename: msg %d: move failed, status=0x%08x",n,status,0,0);
       mm_log(tmp,ERROR);
     }

   chkstat("mail$message_end",   0, mail$message_end   (&msgctx,  NullItemList, NullItemList));
   chkstat("mail$mailfile_close", 0, mail$mailfile_close(&context, NullItemList, NullItemList));
   chkstat("mail$mailfile_end",   0, mail$mailfile_end  (&context, NullItemList, NullItemList));

   __Delete_New_Mailbox(srcmailfile,srcfolder);
   DBGPRINT(DBG$EXIT,"vms_mail_rename: exit",0,0,0,0);
   return T;
}

/* VMS/MAIL mail open - open a "mailbox" which is a folder in VMS/MAIL.
 * Accepts: stream to open
 * Returns: stream on success, NIL on failure
 */

MAILSTREAM *vms_mail_open (MAILSTREAM *stream)
{
  int MailContext=0, MessageContext=0;
  char mailfile[MAILFILESIZE], folder[FOLDERSIZE];
  int inbox, status;
  long i;

					/* return prototype for OP_PROTOTYPE call */
  if (!stream) return &vmsmailproto;


  DBGPRINT(DBG$ENTRY,"vms_mail_open: Entered, mailbox='%s'",stream->mailbox,0,0,0);



/*
  if (!__folder_exists(stream->mailbox, &MailContext, 0, mailfile, folder, &inbox))
    {
      DBGPRINT(DBG$EXIT,"vms_mail_open: folder not found",0,0,0,0);
      mm_log("no such mailbox",ERROR);
      return NIL;
    }
*/
					/* close old file if stream being recycled */
  if (LOCAL) {
    vms_mail_close (stream);		/* dump and save the changes */
    stream->dtb = &vmsmaildriver;	/* reattach this driver */
    mail_free_cache (stream);		/* clean up cache */
  }


					/* Create mailfile context */
  __parse_mailbox_name(stream->mailbox, mailfile, folder, &inbox);
  if ((MailContext = __create_mail_context(mailfile)) == 0)
   {
      DBGPRINT(DBG$EXIT,"vms_mail_open: exit, status = 0 : can't create mail context", 0, 0, 0, 0);
      mm_log("cannot open mailfile",ERROR);
      return NIL;
   }
					/* Create folder and message context */
  if ((MessageContext = __create_msg_context(&MailContext)) == 0)
   {
      __vms_mail_close(&MailContext, &MessageContext);
      DBGPRINT(DBG$EXIT,"vms_mail_open: exit, status = 0 : can't create msg context",0,0,0,0);
      mm_log("cannot open folder", ERROR);
      return NIL;
   }


  stream->dtb = &vmsmaildriver;		/* reattach this driver */
  stream->local = fs_get (sizeof (VMSMAILLOCAL));


					/* Save VMS format of the mailbox */
  strcpy(LOCAL->mailfile,mailfile);
  strcpy(LOCAL->folder,folder);

					/* Note if it is an INBOX */
  LOCAL->inbox=inbox;

					/* Save the mail and message contexts */
  LOCAL->mail_context = MailContext;
  LOCAL->msg_context  = MessageContext;


					/* Provide buffer space */
  LOCAL->buflen = INITTEXTSIZE;
  LOCAL->buf = (char *) fs_get (INITTEXTSIZE + 1);


					/* force readonly if bboard */
  if (*stream->mailbox == '*') stream->rdonly = T;
  stream->sequence++;			/* bump sequence number */


					/* parse mailbox */

  LOCAL->pinged = 0;			/* not 'ping'ed yet! */
  stream->nmsgs = stream->recent = 0;
  if (vms_mail_ping (stream) && !stream->nmsgs)
	mm_log ("Mailbox is empty",(long) NIL);

					/* Initialize the local message cache */
  __init_local_cache(stream);

					/* All elts have valid flags ??? */
  for (i = 1; i < stream->nmsgs; i++)
    mail_elt (stream,i)->valid = T;

					/* Notify application of msg count */
  mail_exists(stream,stream->nmsgs);

  DBGPRINT(DBG$EXIT,"vms_mail_open: exiting",0,0,0,0);
  return stream;			/* return stream to caller */

}

/* VMS/MAIL mail close
 * Accepts: MAIL stream
 */

void vms_mail_close (MAILSTREAM *stream)
{
  DBGPRINT(DBG$ENTRY,"vms_mail_close: entered, mailbox='%s'", stream->mailbox,0,0,0);

					/* if the stream is open, close it */
  if (stream)
    {
      DBGPRINT(DBG$SHOWMORE,"vms_mail_close: closing stream",0,0,0,0);
      if (LOCAL)
        {
          __file_inbox_messages(stream);		/* auto-file read msg*/
          __clear_local_cache(stream);			/* clear cache */
          __vms_mail_close(&LOCAL->mail_context,&LOCAL->msg_context);	/* close file */

							/* return local data*/
          if (LOCAL->buf) fs_give ((void **) &LOCAL->buf);
          fs_give ((void **) &stream->local);
        }
      stream->dtb = NIL;				/* log out the DTB */
    }

  DBGPRINT(DBG$EXIT,"vms_mail_close: exiting",0,0,0,0);
}

/* VMS/MAIL mail fetch fast information
 * Accepts: MAIL stream
 *	    sequence
 */

void vms_mail_fetchfast (MAILSTREAM *stream, char *sequence)
{
  long i;

  DBGPRINT(DBG$ENTRY,"vms_mail_fetchfast: mailbox='%s', sequence='%s'", stream->mailbox, sequence, 0, 0);

				/* scan the sequence */
  if (mail_sequence (stream,sequence))
    for (i = 1; i <= stream->nmsgs; i++)
      if (mail_elt (stream,i)->sequence)
         vms_mail_fetchstructure(stream,i,NIL);

  DBGPRINT(DBG$EXIT,"vms_mail_fetchfast: exit",0,0,0,0);
  return;
}


/* VMS/MAIL mail fetch flags
 * Accepts: MAIL stream
 *	    sequence
 */

void vms_mail_fetchflags (MAILSTREAM *stream, char *sequence)
{
  long i;

  DBGPRINT(DBG$ENTRY,"vms_mail_fetchflags: mailbox='%s', sequence='%s'", stream->mailbox, sequence, 0, 0);

				/* ping mailbox, get new status for messages */
  if (mail_sequence (stream,sequence))
    for (i = 1; i <= stream->nmsgs; i++)
      if (mail_elt (stream,i)->sequence) {
	MESSAGECACHE *elt = mail_elt (stream,i);
        FILECACHE    *fc  = &LOCAL->msgs[i-1];
	elt->deleted  = ((fc->flags & fDELETED)  ? T : NIL);
	elt->seen     = ((fc->flags & fSEEN)     ? T : NIL);
	elt->flagged  = ((fc->flags & fFLAGGED)  ? T : NIL);
	elt->answered = ((fc->flags & fANSWERED) ? T : NIL);
     }
  DBGPRINT(DBG$EXIT,"vms_mail_fetchflags: exit",0,0,0,0);
}

/* VMS/MAIL mail fetch structure
 * Accepts: MAIL stream
 *	    message # to fetch
 *	    pointer to return body
 * Returns: envelope of this message, body returned in body value
 *
 * Fetches the "fast" information as well
 */
ENVELOPE *vms_mail_fetchstructure (MAILSTREAM *stream, long msgno, BODY **body)
{
  FILECACHE *fc = &LOCAL->msgs[msgno-1];
  LONGCACHE *lelt;
  ENVELOPE **env;
  BODY **b;
  STRING bs;

  DBGPRINT(DBG$ENTRY,"vms_mail_fetchstructure: mailbox='%s', msgno='%d'", stream->mailbox, msgno, 0, 0);
  if (stream->scache) {		/* short cache */
    if (msgno != stream->msgno){/* flush old poop if a different message */
      mail_free_envelope (&stream->env);
      mail_free_body (&stream->body);
    }
    stream->msgno = msgno;
    env = &stream->env;		/* get pointers to envelope and body */
    b = &stream->body;
  }
  else {			/* long cache */
    lelt = mail_lelt (stream,msgno);
    env = &lelt->env;		/* get pointers to envelope and body */
    b = &lelt->body;
  }
  if ((body && !*b) || !*env) {	/* have the poop we need? */
    mail_free_envelope (env);	/* flush old envelope and body */
    mail_free_body (b);

				/* Fetch message headers and body */
    __read_message(stream,msgno);

    INIT (&bs,mail_string,(void *) fc->text,fc->size);
				/* parse envelope and body */
    rfc822_parse_msg (env,body ? b : NIL,fc->header,fc->headersize,&bs,mylocalhost(),LOCAL->buf);
    (lelt->elt).rfc822_size = fc->headersize + fc->size;       
  }

  DBGPRINT(DBG$EXIT,"vms_mail_fetchstructure: exit",0,0,0,0);
  if (body) *body = *b;		/* return the body */
  return *env;			/* return the envelope */
}

/* VMS/MAIL mail fetch message header
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message header in RFC822 format
 */

char *vms_mail_fetchheader (MAILSTREAM *stream,long msgno)
{
  FILECACHE *fc = &LOCAL->msgs[msgno-1];

  DBGPRINT(DBG$ENTRY,"vms_mail_fetchheader: mailbox='%s', msgno='%d'", stream->mailbox, msgno, 0, 0);

  __read_message(stream,msgno);
  strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,fc->header,fc->headersize);

  DBGPRINT(DBG$EXIT,"vms_mail_fetchheader: exit",0,0,0,0);
  return LOCAL->buf;
}




/* VMS/MAIL mail fetch message text (only)
	body only;
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message text in RFC822 format
 */

char *vms_mail_fetchtext (MAILSTREAM *stream, long msgno)
{
  FILECACHE    *fc  = &LOCAL->msgs[msgno-1];
  MESSAGECACHE *elt = mail_elt(stream,msgno);

  DBGPRINT(DBG$ENTRY,"vms_mail_fetchtext: mailbox='%s', msgno='%d'", stream->mailbox, msgno, 0, 0);
				/* get to text position */
  __read_message(stream,msgno);
				/* copy the string */
  strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,fc->text,fc->size);

				/* recalculate status */
  if (!elt->seen)
    { elt->seen = T; __update_message_status (stream,msgno,T); }

  DBGPRINT(DBG$EXIT,"vms_mail_fetchtext: exit",0,0,0,0);
  return LOCAL->buf;
}

/* VMS/MAIL fetch message body as a structure
 * Accepts: Mail stream
 *	    message # to fetch
 *	    section specifier
 *	    pointer to length
 * Returns: pointer to section of message body
 */
char *vms_mail_fetchbody (MAILSTREAM *stream, long m, char *s, unsigned long *len)
{
  BODY *b;
  PART *pt;
  char *t;
  ENVELOPE *env;
  unsigned long i;
  unsigned long offset = 0;
  MESSAGECACHE *elt = mail_elt (stream,m);
  FILECACHE    *fc  = &LOCAL->msgs[m-1];

  DBGPRINT(DBG$ENTRY,"vms_mail_fetchbody: mailbox='%s', msgno='%d', section='%s'", stream->mailbox, m, s, 0);
				/* make sure have a body */
  env = vms_mail_fetchstructure (stream,m,&b);
  if (!(env && b && s && *s &&
	((i = strtol (s,&s,10)) > 0))) return NIL;
  do {				/* until find desired body part */
				/* multipart content? */
    if (b->type == TYPEMULTIPART) {
      pt = b->contents.part;	/* yes, find desired part */
      while (--i && (pt = pt->next));
      if (!pt) return NIL;	/* bad specifier */
				/* note new body, check valid nesting */
      if (((b = &pt->body)->type == TYPEMULTIPART) && !*s) return NIL;
      offset = pt->offset;	/* get new offset */
    }
    else if (i != 1) return NIL;/* otherwise must be section 1 */
				/* need to go down further? */
    if (i = *s) switch (b->type) {
    case TYPEMESSAGE:		/* embedded message, calculate new base */
      offset = b->contents.msg.offset;
      b = b->contents.msg.body;	/* get its body, drop into multipart case */
    case TYPEMULTIPART:		/* multipart, get next section */
      if ((*s++ == '.') && (i = strtol (s,&s,10)) > 0) break;
    default:			/* bogus subpart specification */
      return NIL;
    }
  } while (i);
				/* lose if body bogus */
  if ((!b) || b->type == TYPEMULTIPART) return NIL;

  if (!elt->seen)
    { elt->seen=T;  __update_message_status(stream,m,T); }

				/* move to that place in the data */
  t = fc->text + offset;

  rfc822_contents(&LOCAL->buf,&LOCAL->buflen,len,t,b->size.ibytes,b->encoding);
  DBGPRINT(DBG$EXIT,"vms_mail_fetchbody: exit",0,0,0,0);
  return LOCAL->buf;
}

/* Parse flag list
 * Accepts: MAIL stream
 *	    flag list as a character string
 * Returns: flag command list
 */


short vms_mail_getflags (MAILSTREAM *stream, char *flag)
{
  char tmpbuf[256];
  char *t;
  short f = 0;
  short i,j;
  if (flag && *flag) {		/* no-op if no flag string */
				/* check if a list and make sure valid */
    if ((i = (*flag == '(')) ^ (flag[strlen (flag)-1] == ')')) {
      mm_log ("Bad flag list",ERROR);
      return NIL;
    }
				/* copy the flag string w/o list construct */
    strncpy (tmpbuf,flag+i,(j = strlen (flag) - (2*i)));
    tmpbuf[j] = '\0';
    t = ucase (tmpbuf);		/* uppercase only from now on */

    while (*t) {		/* parse the flags */
      if (*t == '\\') {		/* system flag? */
	switch (*++t) {		/* dispatch based on first character */
	case 'S':		/* possible \Seen flag */
          if (strncmp(t+1, "EEN", 3) == 0) i = fSEEN;
	  t += 4;		/* skip past flag name */
	  break;
	case 'D':		/* possible \Deleted flag */
          if (strncmp(t+1, "ELETED", 6) == 0) i = fDELETED;
	  t += 7;		/* skip past flag name */
	  break;
	case 'F':		/* possible \Flagged flag */
          if (strncmp(t+1, "LAGGED", 6) == 0) i = fFLAGGED;
	  t += 7;		/* skip past flag name */
	  break;
	case 'A':		/* possible \Answered flag */
          if (strncmp(t+1,"NSWERED", 7) == 0) i = fANSWERED;
	  t += 8;		/* skip past flag name */
	  break;
	default:		/* unknown */
	  i = 0;
	  break;
	}
				/* add flag to flags list */
	if (i && ((*t == '\0') || (*t++ == ' '))) f |= i;
	else {			/* bitch about bogus flag */
	  mm_log ("Unknown system flag",ERROR);
	  return NIL;
	}
      }
      else {			/* no user flags yet */
	mm_log ("Unknown flag",ERROR);
	return NIL;
      }
    }
  }
  return f;
}

/* VMS/MAIL mail set flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void vms_mail_setflag (MAILSTREAM *stream, char *sequence, char *flag)
{
  int	i;
  MESSAGECACHE *elt;

  short f = vms_mail_getflags (stream,flag);	/* Parse the list */
  if (!f) return;		/* no-op if no flags to modify */

				/* get sequence and loop on it */
  if (mail_sequence (stream,sequence)) for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence)
      {
        __setflag(stream,i,elt,f);
        __update_message_status(stream,i,T);
      }
}




/* VMS/MAIL mail clear flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void vms_mail_clearflag (MAILSTREAM *stream, char *sequence, char *flag)
{
  int	i;
  MESSAGECACHE *elt;

  short f = vms_mail_getflags (stream,flag);	/* Parse the list */
  if (!f) return;		/* no-op if no flags to modify */
				/* get sequence and loop on it */
  if (mail_sequence (stream,sequence)) for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence)
      {
        __clearflag(stream,i, elt, f);
        __update_message_status(stream,i,T);
      }
}

/* VMS/MAIL mail search for messages
 * Accepts: MAIL stream
 *	    search criteria
 */

void vms_mail_search (MAILSTREAM *stream, char *criteria)
{
  long i,n;
  char *d;
  search_t f;
				/* make sure all messages read in */
  for (i=1; i<=stream->nmsgs; i++)
    vms_mail_fetchstructure(stream, i, NIL);

				/* initially all searched */
  for (i = 1; i <= stream->nmsgs; ++i) mail_elt (stream,i)->searched = T;
				/* get first criterion */
  if (criteria && (criteria = strtok (criteria," "))) {
				/* for each criterion */
    for (; criteria; (criteria = strtok (NIL," "))) {
      f = NIL; d = NIL; n = 0;	/* init then scan the criterion */
      switch (*ucase (criteria)) {
      case 'A':			/* possible ALL, ANSWERED */
	if (!strcmp (criteria+1,"LL")) f = vms_mail_search_all;
	else if (!strcmp (criteria+1,"NSWERED")) f = vms_mail_search_answered;
	break;
      case 'B':			/* possible BCC, BEFORE, BODY */
	if (!strcmp (criteria+1,"CC"))
	  f = vms_mail_search_string (vms_mail_search_bcc,&d,&n);
	else if (!strcmp (criteria+1,"EFORE"))
	  f = vms_mail_search_date (vms_mail_search_before,&n);
	else if (!strcmp (criteria+1,"ODY"))
	  f = vms_mail_search_string (vms_mail_search_body,&d,&n);
	break;
      case 'C':			/* possible CC */
	if (!strcmp (criteria+1,"C"))
	  f = vms_mail_search_string (vms_mail_search_cc,&d,&n);
	break;
      case 'D':			/* possible DELETED */
	if (!strcmp (criteria+1,"ELETED")) f = vms_mail_search_deleted;
	break;
      case 'F':			/* possible FLAGGED, FROM */
	if (!strcmp (criteria+1,"LAGGED")) f = vms_mail_search_flagged;
	else if (!strcmp (criteria+1,"ROM"))
	  f = vms_mail_search_string (vms_mail_search_from,&d,&n);
	break;
      case 'K':			/* possible KEYWORD */
	if (!strcmp (criteria+1,"EYWORD"))
	  f = vms_mail_search_flag (vms_mail_search_keyword,&n,stream);
	break;
      case 'N':			/* possible NEW */
	if (!strcmp (criteria+1,"EW")) f = vms_mail_search_new;
	break;
      case 'O':			/* possible OLD, ON */
	if (!strcmp (criteria+1,"LD")) f = vms_mail_search_old;
	else if (!strcmp (criteria+1,"N"))
	  f = vms_mail_search_date (vms_mail_search_on,&n);
	break;
      case 'R':			/* possible RECENT */
	if (!strcmp (criteria+1,"ECENT")) f = vms_mail_search_recent;
	break;
      case 'S':			/* possible SEEN, SINCE, SUBJECT */
	if (!strcmp (criteria+1,"EEN")) f = vms_mail_search_seen;
	else if (!strcmp (criteria+1,"INCE"))
	  f = vms_mail_search_date (vms_mail_search_since,&n);
	else if (!strcmp (criteria+1,"UBJECT"))
	  f = vms_mail_search_string (vms_mail_search_subject,&d,&n);
	break;
      case 'T':			/* possible TEXT, TO */
	if (!strcmp (criteria+1,"EXT"))
	  f = vms_mail_search_string (vms_mail_search_text,&d,&n);
	else if (!strcmp (criteria+1,"O"))
	  f = vms_mail_search_string (vms_mail_search_to,&d,&n);
	break;
      case 'U':			/* possible UN* */
	if (criteria[1] == 'N') {
	  if (!strcmp (criteria+2,"ANSWERED")) f = vms_mail_search_unanswered;
	  else if (!strcmp (criteria+2,"DELETED")) f = vms_mail_search_undeleted;
	  else if (!strcmp (criteria+2,"FLAGGED")) f = vms_mail_search_unflagged;
	  else if (!strcmp (criteria+2,"KEYWORD"))
	    f = vms_mail_search_flag (vms_mail_search_unkeyword,&n,stream);
	  else if (!strcmp (criteria+2,"SEEN")) f = vms_mail_search_unseen;
	}
	break;
      default:			/* we will barf below */
	break;
      }
      if (!f) {			/* if can't determine any criteria */
	sprintf (LOCAL->buf,"Unknown search criterion: %.80s",criteria);
	mm_log (LOCAL->buf,ERROR);
	return;
      }
				/* run the search criterion */
      for (i = 1; i <= stream->nmsgs; ++i)
	if (mail_elt (stream,i)->searched && !(*f) (stream,i,d,n))
	  mail_elt (stream,i)->searched = NIL;
    }
				/* report search results to main program */
    for (i = 1; i <= stream->nmsgs; ++i)
      if (mail_elt (stream,i)->searched) mail_searched (stream,i);
  }
}

/* VMS/MAIL mail ping mailbox - find how many messages are in this folder.
 * Do it only when opening the folder.
 *
 * Accepts: MAIL stream
 * Returns: T if stream still alive, NIL if not
 */
long vms_mail_ping (MAILSTREAM *stream)
{
  static unsigned int MessagesCount;
  int status;

  item FolderItem[] = {				/* Folder's name */
	{ 0, MAIL$_MESSAGE_FOLDER, NULL, NULL },/* Must be the first item!!! */
	{ 0, MAIL$_NOSIGNAL,       NULL, NULL},
	{ 0, 0, NULL, NULL }};

  item MessagesCountItem[] = {			/* No of messages */
	{ sizeof(MessagesCount), MAIL$_MESSAGE_SELECTED,&MessagesCount, NULL },
	{ 0, 0, NULL, NULL }};

  DBGPRINT(DBG$ENTRY,"vms_mail_ping: entered, mailbox='%s'",stream->mailbox,0,0,0);

  if (!LOCAL->pinged) 		/* Not checked yet ? */
    {
      MessagesCount = 0;
      FolderItem[0].buffer = LOCAL->folder;
      FolderItem[0].length = strlen(LOCAL->folder);
      status = mail$message_select(&LOCAL->msg_context, FolderItem, MessagesCountItem);
      chkstat("mail$message_select", 0, status);
      LOCAL->pinged = 1;
      stream->nmsgs = MessagesCount;

				/* Check list of 'new' folders if not found */
      if (status == MAIL$_NOTEXIST)
        {
          DBGPRINT(DBG$EXIT,"vms_mail_ping: exit, messages='%u'", stream->nmsgs,0,0,0);
          return __search_new_mailboxes(LOCAL->mailfile, LOCAL->folder);
        }
    }

  DBGPRINT(DBG$EXIT,"vms_mail_ping: exit, messages='%u'", stream->nmsgs,0,0,0);
  return T;		/* local stream is always alive */
}

/* VMS/MAIL mail check mailbox (too)
	reparses status too;
 * Accepts: MAIL stream
 */

void vms_mail_check (MAILSTREAM *stream)
{
  mm_log ("Check completed",(long) NIL);
}

/* VMS/MAIL mail expunge mailbox - Delete the messages that are flagged as
 * deleted.
 * Accepts: MAIL stream
 * 
 *
 * Comments:
 *      Expunge may be called twice for the same stream. Once just prior
 *      to calling the close operation and again, if an INBOX, as part of the
 *      close operation. The latter is used to delete messages that have been
 *      auto-filed to another folder.
 * 
 *	To avoid problems in deleting messages twice, we use a separate flag
 *	WAS_DELETED to indicate if we deleted the message the first time
 *	around. That way, we know not to delete it again.
 */

void vms_mail_expunge (MAILSTREAM *stream)
{
  static int MessageNum;		/* Message number to delete */
  int status;
  long i, f;
  long MessagesDeleted = 0;		/* How many messages were deleted */

  item DeleteMessageItem[] = {
	{ sizeof(MessageNum), MAIL$_MESSAGE_ID, &MessageNum, NULL },
	{ 0, 0, NULL, NULL }};

  
  DBGPRINT(DBG$ENTRY,"vms_mail_expunge: entered, mailbox='%s'", stream->mailbox,0,0,0);
  for (i = 1; i <= stream->nmsgs; ++i)
    {
      f = LOCAL->msgs[i-1].flags;
      if( (f & fDELETED)  &&  !(f & fWAS_DELETED) )
	{
	  MessageNum = i;	/* Assign it at the descriptor */
	  status = mail$message_delete(&LOCAL->msg_context, &DeleteMessageItem, NullItemList);
	  if (chkstat("mail$message_delete", 0, status))
            {
              DBGPRINT(DBG$SHOWMORE,"vms_mail_expunge: Message '%d' deleted", i, 0, 0, 0);
	      LOCAL->msgs[i - 1].flags |= fWAS_DELETED;
              MessagesDeleted++;	/* Count ALL deleted messages */
            }
	}
    }


  if (MessagesDeleted)			/* If any messages were removed */
    {
     char expmess[256];
     sprintf(expmess,
        "Expunged %d message%s", MessagesDeleted, MessagesDeleted!=1?"s":"");
     mm_log(expmess,NIL);     
    }

  DBGPRINT(DBG$EXIT,"vms_mail_expunge: exit, messages expunged ='%d'", MessagesDeleted,0,0,0);
}

/* VMS/MAIL mail copy message(s)
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 * Returns: T if success, NIL if failed
 */

long vms_mail_copy (MAILSTREAM *stream, char *sequence, char *mailbox)
{
				/* copy the messages */
  return (mail_sequence (stream,sequence)) ?
    vms_mail_copy_messages (stream,mailbox) : NIL;
}


/* VMS/MAIL mail move message(s)
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 * Returns: T if success, NIL if failed
 */

long vms_mail_move (MAILSTREAM *stream, char *sequence, char *mailbox)
{
  long i;
  MESSAGECACHE *elt;
  if (!(mail_sequence (stream,sequence) &&
	vms_mail_copy_messages (stream,mailbox))) return NIL;

				/* delete all requested messages */
  for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence) {
      elt->deleted = T;		/* mark message deleted */
				/* recalculate status */
      __update_message_status (stream,i,NIL);
    }
				/* make sure the update takes */
  return LONGT;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Copy message(s) from the named sequential mail file into another
	mailfile/folder

    Inputs:
	seqfile		Name of a sequential mail file containing source
	msg_count	Number of messages to copy
	mailfile	Name of destination indexed mail file
	folder		Name of destination mail folder

    Outputs:
	NONE

    Return Status:
	T		Copied OK
	NIL		Some problem during copy

    Author:
	Thanks to Steve Pulley of Digital for basic idea and code, November 1996
	Modified by Andy Harper, Kings College London, November 1996.

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __copy_sequential_mail_messages(char *seqfile, int msg_count, char *mailfile, char *folder)
  {
    static int id, newfile, newfold;
    int file_ctx=0, msg_context=0;
    int status;

    item copy_in[]   = {
	{sizeof(id),       MAIL$_MESSAGE_ID,               &id,        NULL},
	{strlen(mailfile), MAIL$_MESSAGE_FILENAME,         mailfile,   NULL},
	{strlen(folder),   MAIL$_MESSAGE_FOLDER,           folder,     NULL},
	{0,                MAIL$_NOSIGNAL,                 NULL,       NULL},
	{0,                0,                              NULL,       NULL}};

    item copy_out[]  = {
	{sizeof(newfile),  MAIL$_MESSAGE_FILE_CREATED,     &newfile,   NULL},
	{sizeof(newfold),  MAIL$_MESSAGE_FOLDER_CREATED,   &newfold,   NULL},
	{0,                0,                              NULL,       NULL}};

    DBGPRINT(DBG$ENTRY,"__copy_sequential_mail_messages: entered, seq file='%s', mailfile='%s', folder='%s'\n", seqfile, mailfile, folder, 0);

    DBGPRINT(DBG$SHOWMORE,"__copy_sequential_mail_messages: Begin mail context",0,0,0,0);
    if (file_ctx = __create_mail_context(seqfile))
      if (msg_context = __create_msg_context(&file_ctx))
        {
          DBGPRINT(DBG$SHOWMORE,"__copy_sequential_mail_messages: Begin message copy",0,0,0,0);
          for (id = 1; id <= msg_count; id++)
            {
              DBGPRINT(DBG$SHOWMORE,"__copy_sequential_mail_messages: copying message %d", id,0,0,0);
              status = mail$message_copy(&msg_context, copy_in, copy_out);
              if ( !chkstat("mail$message_copy", 0, status) )
                {
                  DBGPRINT(DBG$EXIT,"__copy_sequential_mail_messages: exit, copy failed, status=0",0,0,0,0);
                  __vms_mail_close(&file_ctx,&msg_context);
                  return NIL;
                }

              if (newfile)
                {
                  DBGPRINT(DBG$SHOWMORE,"__copy_sequential_mail_messages: created mail file",0,0,0,0);
                }

              if (newfold)
                {
                  DBGPRINT(DBG$SHOWMORE,"__copy_sequential_mail_messages: created folder",0,0,0,0);
                }
            }
        }

    __vms_mail_close(&file_ctx, &msg_context);

    DBGPRINT(DBG$EXIT,"__copy_sequential_mail_messages: exit, status=T",0,0,0,0);
    return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Append a message stored in a STRING structure to a nominated mailbox

   Inputs:
	stream		A mail stream
	mailbox		The mailbox to which is to be appended
	flags		Pointer to list of flags to be set (can be NULL!)
	date		Pointer to date of message         (can be NULL!)
	message		The STRING structure holding the message

   Outputs:
	NONE

   Return Value:
	T		If message appended successfully
	NIL		If not

   Comments:
	'append' in VMS MAIL terms actually means copying the message into
	a particular folder. Its position will be based on the date.

	Appending a message to an arbitrary folder in VMS MAIL is non-trivial.
	Essentially, what we do is to write a temporary file in the form of a
	VMS sequential mail file, then use the mail copy function to get it
	into the required folder.

	The format of a sequential mail file has to be fairly precise but is
        not obviously documented. I've worked it out by trial and error to be
	this (though of course it may change in the future!):

	    From:<tab>[<from-address><spaces>[<personal-name><spaces>]]<date/time>
	    To:<tab>[<to-address>]
	    CC:<tab>[<cc-address>]
	    Subj:<tab>[<subject text>]
	    <blank line>
	    <text body>

	<..> items are data to be filled in
	[..] items are optional

	The data in the string structure is simply the text as received
	(normally the concatenation of the RFC822 format headers and the text
	body) so the headers must be parsed to extract the above information.

        The message text may contain '\r' characters. They are stripped out
	otherwise they confuse mail's header parsing. The '\n' is left alone
	though.

   Limitations:
	Currently, I can see no way to action the flags setting as, once the
        message has been copied into the mail folder, you have no means to
        locate it in order to use the message_modify function. Nor can you
	supply the flags setting at the time of the copy. This may be a 
	permanent restriction.

   Author:
	Andy Harper, Kings College London, January 1997
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
long vms_mail_append (MAILSTREAM *stream, char *mailbox, char *flags, char *date, STRING *message)
  {
    static char *months[] = {
	"JAN","FEB","MAR","APR","MAY","JUN","JUL","AUG","SEP","OCT","NOV","DEC"
	};
    static char *filepat[] = {
	"SYS$SCRATCH:SEQMAIL%08x.TMP",
	"SYS$LOGIN:SEQMAIL%08x.TMP",
	"SYS$DISK:[]SEQMAIL%08x.TMP"
			  };
    static int        maxfiles     = sizeof(filepat) / sizeof(filepat[0]);
    long int          size         = SIZE(message);

    FILE              *fp;		/* for sequential mail file */
    ENVELOPE          *env;
    ADDRESS           *ap;
    MESSAGECACHE      elt;
    char              filename[MAILFILESIZE],mailfile[MAILFILESIZE],folder[FOLDERSIZE];
    char              address[256],datebuffer[256];
    char              c,*hdr,*ptr,*line,*tmp;
    int               hdrsize,inbox,i;
    short             f;

    DBGPRINT(DBG$ENTRY,"vms_mail_append: entered, mailbox='%s', flags='%s', size='%ld', date='%s'", mailbox ? mailbox : "(null)", flags ? flags : "(null)", size, date ? date : "(null)");

					/* If no date, use today's date */
    if (!date)
      { rfc822_date(datebuffer); date = datebuffer; }

					/* Get fields of date */
    DBGPRINT(DBG$ENTRY,"vms_mail_append: date='%s'", date, 0, 0, 0);
    if (!mail_parse_date(&elt,date))
     {
       DBGPRINT(DBG$SHOWMORE,"vms_mail_append: date parse of '%s' failed",date,0,0,0);
       mm_log("unparseable date field",ERROR);
       return NIL;
     }
					/* Create VMS format date */
    sprintf(datebuffer,"%d-%s-%04d %2d:%02d:%02d",
      elt.day,months[elt.month-1],elt.year+BASEYEAR,elt.hours,elt.minutes,elt.seconds);


    DBGPRINT(DBG$SHOWMORE,"vms_mail_append: VMS format date ='%s'", datebuffer, 0, 0, 0);

					/* Get the flags -- (NOT IMPLEMENTED!) */
/*    f = vms_mail_getflags(stream,flags); */


					/* parse mailbox name */
    if (!__folder_exists(mailbox,NULL,1,mailfile,folder,&inbox))
      {
        char buf[256];
        sprintf(buf,"no such mailbox '%s'",mailbox);
        mm_log(buf,ERROR);
        return NIL;
      }

					/* Init to scan message */
    line=ptr=hdr = (char *) fs_get(size+1);


					/* Accumulate header until blank line */
    while (size--)
      if ((c=SNX(message)) != '\r')	/* Ignore \r characters */
        if ((*ptr++ = c) == '\n')
          {
            DBGPRINT(DBG$SHOWMORE,"vms_mail_append: line '%.*s'", ptr-line, line, 0, 0);
            if (ptr-line <= 1) break;	/* If blank line, stop here */
            line=ptr;			/* .. else note start of new line */
           }

					/* Parse the RFC822 headers */
    DBGPRINT(DBG$SHOWMORE,"vms_mail_append: parsing headers", 0, 0, 0, 0);
    hdrsize = ptr-hdr;
    tmp = (char *) fs_get(hdrsize + 1);
    rfc822_parse_msg(&env, NULL, hdr, hdrsize, NULL, mylocalhost(), tmp);
    fs_give((void **) &tmp);

					/* Try to open a scratch file */
    DBGPRINT(DBG$SHOWMORE,"vms_mail_append: opening a scratch file (%d attempts)", maxfiles, 0, 0, 0);
    for (i=0; i<maxfiles; i++)
      {
        sprintf(filename, filepat[i], getpid());
        DBGPRINT(DBG$SHOWMORE,"vms_mail_append: trying '%s'", filename, 0, 0, 0);
        if ((fp=fopen(filename, "w", "rfm=var","rat=cr")) != NULL) break;
        DBGPRINT(DBG$SHOWMORE,"vms_mail_append: failed to open '%s'", filename, 0, 0, 0);
      }

    if (fp == NULL)
      { DBGPRINT(DBG$SHOWMORE,"vms_mail_append: exit, status=0 : cant write seq mail file",0,0,0,0);
        mm_log("cannot write sequential mail file",ERROR);
        fs_give((void **)&hdr);
        return NIL;
      }

					/* Determine transport prefix */
     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: Looking for transport prefix",0,0,0,0);
     if ((tmp=__transport_name()) != NULL)
       {
         sprintf(address,"%s%%%%\"%%s@%%s\"%%s",tmp);
         DBGPRINT(DBG$SHOWMORE,"vms_mail_append: prefix='%s', address format='%s'",tmp,address,0,0);
       }
     else
       {
         strcpy(address,"%s@%s%s");
         DBGPRINT(DBG$SHOWMORE,"vms_mail_append: no prefix, address format '%s'", address,0,0,0);
       }

					/* Write the VMS headers */
     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: writing VMS headers to %s",filename,0,0,0);
     fputc('\f',fp);
     fputc('\n',fp);

     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: processing from: header",0,0,0,0);
     fprintf(fp,"From:\t");
     for (ap=env->from; ap ; ap=ap->next)
      { DBGPRINT(DBG$SHOWMORE,"vms_mail_append: FROM: <%s@%s>", ap->mailbox,ap->host,0,0);
        fprintf(fp, address, ap->mailbox, ap->host, ap->next ?  "," : "");}
     fprintf(fp," %s\n", datebuffer);

     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: processing to: header",0,0,0,0);
     fprintf(fp,"To:\t");
     for (ap=env->to; ap ; ap=ap->next)
      { DBGPRINT(DBG$SHOWMORE,"vms_mail_append: TO:   <%s@%s>", ap->mailbox,ap->host,0,0);
        fprintf(fp, address, ap->mailbox, ap->host, ap->next ?  "," : "");}
     fputc('\n',fp);

     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: processing cc: header",0,0,0,0);
     fprintf(fp,"CC:\t");
     for (ap=env->cc; ap ; ap=ap->next)
      { DBGPRINT(DBG$SHOWMORE,"vms_mail_append: CC:   <%s@%s>", ap->mailbox,ap->host,0,0);
        fprintf(fp, address, ap->mailbox, ap->host, ap->next ?  "," : "");}
     fputc('\n',fp);

     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: processing subj: header",0,0,0,0);
     fprintf(fp,"Subj:\t%s\n", env->subject ? env->subject : "");
     fputc('\n',fp);

					/* Free up the envelope */
     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: free envelope structure",0,0,0,0);
     mail_free_envelope(&env);

					/* Write the RFC822 headers */
     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: writing RFC822 headers (size='%d')\n%.*s",hdrsize,hdrsize,hdr,0);
     for (ptr=hdr; hdrsize-- > 0; ptr++)
       fputc(*ptr,fp);
     fs_give((void **)&hdr);

					/* Copy and write the body */
     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: writing message text (size=%d)",size,0,0,0);
     while (size-- > 0)
       {c=SNX(message); if (c != '\r') fputc(c, fp);}


     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: closing sequential mail file",0,0,0,0);
     fclose(fp);

					/* Add message to folder */
     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: copying message into '%s'",mailbox,0,0,0);
     if (!__copy_sequential_mail_messages(filename, 1, mailfile, folder))
       { DBGPRINT(DBG$SHOWMORE,"vms_mail_append: append to folder failed",0,0,0,0);
         mm_log("append to folder failed",ERROR);
         remove(filename);
         return NIL;
       }

     DBGPRINT(DBG$SHOWMORE,"vms_mail_append: deleting temporary file '%s'",filename,0,0,0);
     remove(filename);

					/* Return appropriate status */
    DBGPRINT(DBG$EXIT,"vms_mail_append: exit,status=1",0,0,0,0);
    return T;
}

/* VMS/MAIL garbage collect stream
 * Accepts: Mail stream
 *	    garbage collection flags
 */

void vms_mail_gc (MAILSTREAM *stream, long gcflags)
{
  /* nothing here for now */
}

/* Search support routines
 * Accepts: MAIL stream
 *	    message number
 *	    pointer to additional data
 * Returns: T if search matches, else NIL
 */


char vms_mail_search_all (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return T;			/* ALL always succeeds */
}


char vms_mail_search_answered (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->answered ? T : NIL;
}


char vms_mail_search_deleted (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->deleted ? T : NIL;
}


char vms_mail_search_flagged (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->flagged ? T : NIL;
}


char vms_mail_search_keyword (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->user_flags & n ? T : NIL;
}


char vms_mail_search_new (MAILSTREAM *stream, long msgno, char *d, long n)
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (elt->recent && !elt->seen) ? T : NIL;
}



char vms_mail_search_old (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->recent ? NIL : T;
}


char vms_mail_search_recent (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->recent ? T : NIL;
}


char vms_mail_search_seen (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->seen ? T : NIL;
}


char vms_mail_search_unanswered (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->answered ? NIL : T;
}


char vms_mail_search_undeleted (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->deleted ? NIL : T;
}


char vms_mail_search_unflagged (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->flagged ? NIL : T;
}


char vms_mail_search_unkeyword (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->user_flags & n ? NIL : T;
}


char vms_mail_search_unseen (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->seen ? NIL : T;
}



char vms_mail_search_before (MAILSTREAM *stream, long msgno, char *d, long n)
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) ((long)((elt->year << 9) + (elt->month << 5) + elt->day) < n);
}


char vms_mail_search_on (MAILSTREAM *stream, long msgno, char *d, long n)
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) (((elt->year << 9) + (elt->month << 5) + elt->day) == n);
}


char vms_mail_search_since (MAILSTREAM *stream, long msgno, char *d, long n)
{
				/* everybody interprets "since" as .GE. */
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) ((long)((elt->year << 9) + (elt->month << 5) + elt->day) >= n);
}

char vms_mail_search_body (MAILSTREAM *stream, long msgno, char *d, long n)
{
  char ret;
  char *s = vms_mail_fetchtext (stream,msgno);	/* get the text in LOCAL->buf */

  ret = search (s,strlen(s),d,n);/* do the search */
  return ret;			/* return search value */
}

char vms_mail_search_subject (MAILSTREAM *stream, long msgno, char *d, long n)
{
  char *s = vms_mail_fetchstructure (stream,msgno,NIL)->subject;
  return s ? search (s,(long) strlen (s),d,n) : NIL;
}


char vms_mail_search_text (MAILSTREAM *stream, long msgno, char *d, long n)
{
  char ret;
  char *s = vms_mail_fetchheader (stream,msgno);	/* Get the header in LOCAL->buf */

  ret = search (s,strlen(s),d,n);  /* First, search at the header */
  if(!ret)	/* Not found at header, search at the body */
	  ret = vms_mail_search_body (stream,msgno,d,n);
  return ret;			/* return search value */
}



char vms_mail_search_bcc (MAILSTREAM *stream, long msgno, char *d, long n)
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->bcc;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_cc (MAILSTREAM *stream, long msgno, char *d, long n)
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->cc;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_from (MAILSTREAM *stream, long msgno, char *d, long n)
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->from;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_to (MAILSTREAM *stream, long msgno, char *d, long n)
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->to;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}

/* Search parsers */


/* Parse a date
 * Accepts: function to return
 *	    pointer to date integer to return
 * Returns: function to return
 */

search_t vms_mail_search_date (search_t f, long *n)
{
  long i;
  char *s;
  MESSAGECACHE elt;
				/* parse the date and return fn if OK */
  return (vms_mail_search_string (f,&s,&i) && mail_parse_date (&elt,s) &&
	  (*n = (elt.year << 9) + (elt.month << 5) + elt.day)) ? f : NIL;
}



/* Parse a flag
 * Accepts: function to return
 *	    pointer to keyword integer to return
 *	    MAIL stream
 * Returns: function to return
 */

search_t vms_mail_search_flag (search_t f, long *n,MAILSTREAM *stream)
{
  short i;
  char *s,*t;
  if (t = strtok (NIL," ")) {	/* get a keyword */
    ucase (t);			/* get uppercase form of flag */
    for (i = 0; i < NUSERFLAGS && (s = stream->user_flags[i]); ++i)
      if (!strcmp (t,ucase (strcpy (LOCAL->buf,s))) && (*n = 1 << i)) return f;
  }
  return NIL;			/* couldn't find keyword */
}

/* Parse a string
 * Accepts: function to return
 *	    pointer to string to return
 *	    pointer to string length to return
 * Returns: function to return
 */


search_t vms_mail_search_string (search_t f,char **d, long *n)
{
  char *end = " ";
  char *c = strtok (NIL,"");	/* remainder of criteria */
  if (!c) return NIL;		/* missing argument */
  switch (*c) {			/* see what the argument is */
  case '{':			/* literal string */
    *n = strtol (c+1,d,10);	/* get its length */
    if ((*(*d)++ == '}') && (*(*d)++ == '\015') && (*(*d)++ == '\012') &&
	(!(*(c = *d + *n)) || (*c == ' '))) {
      char e = *--c;
      *c = DELIM;		/* make sure not a space */
      strtok (c," ");		/* reset the strtok mechanism */
      *c = e;			/* put character back */
      break;
    }
  case '\0':			/* catch bogons */
  case ' ':
    return NIL;
  case '"':			/* quoted string */
    if (strchr (c+1,'"')) end = "\"";
    else return NIL;
  default:			/* atomic string */
    if (*d = strtok (c,end)) *n = strlen (*d);
    else return NIL;
    break;
  }
  return f;
}

static unsigned int SendContext = 0;
static int MessageSentCount;		/* Set by the error/success routine */


/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Purpose:
	Check error status, cancel the current send and return pointer to
	suitable error message.

     Author:
	Andy Harper, Kings College London, Oct 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *chksndstat(char *message, int status)
  {
     static char errmess[MAXRECORDSIZE];

     DBGPRINT(DBG$SYSSTATUS,"%s: exit, status=0x%08x", message, status, 0, 0);

     if ((status & 0x1) == 0)
       {
         sprintf(errmess,"%s: failed with status 0x%08x", message, status);
         if (SendContext != 0) mail$send_end(&SendContext, NullItemList, NullItemList);
         return errmess;
       }

     return NULL;
  }




/*
 | this routine is called only if the message was sent ok. We count the number
 | of messages sent ok.
 */
static int success_send_result(item *recipient, int *sigarray, int userdata)
  {
    MessageSentCount++;	/* OK */
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 | Get the message's text filename, the addresses and send it using the
 | selected mail protocol.
 | Returns NULL in case of success, or a pointer to the error message.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
char *vms_mail_send(char *filename, ADDRESS *from, ADDRESS *to, ADDRESS *cc, char *subject)
  {
    static char line[MAXRECORDSIZE];
    static short int typesu = MAIL$_TO, typecc = MAIL$_CC;
    ADDRESS *ap;
    FILE *ifd;
    int status, AddressesCount;
    char *err,*p, *MailProtocolUsed;

    item ToItem[] = {
	{ 0,              MAIL$_SEND_USERNAME,      line,               NULL},
	{ sizeof(typesu), MAIL$_SEND_USERNAME_TYPE, &typesu,            NULL},
	{ 0, 0, NULL, NULL }};

    item CcItem[] = {
	{ 0,              MAIL$_SEND_USERNAME,      line,                NULL},
	{ sizeof(typecc), MAIL$_SEND_USERNAME_TYPE, &typecc,             NULL},
	{ 0, 0, NULL, NULL }};

    item FromItem[] = {
	{ 0,              MAIL$_SEND_FROM_LINE,     line,                NULL},
	{ 0, 0, NULL, NULL }};

    item SubjectItem[] = {
	{ 0,              MAIL$_SEND_SUBJECT,       line,                NULL},
	{ 0, 0, NULL, NULL }};

    item RecordItem[] = {
	{ 0,              MAIL$_SEND_RECORD,        line,                NULL},
	{ 0, 0, NULL, NULL }};

    item FileItem[] = {
	{ 0,              MAIL$_SEND_FILENAME,      line,                NULL},
	{ 0, 0, NULL, NULL }};

    item SendItem[] = {
	{ sizeof (int *), MAIL$_SEND_SUCCESS_ENTRY, success_send_result, NULL},
	{ 0,              MAIL$_NOSIGNAL,           NULL,                NULL},
	{ 0, 0, NULL, NULL }};

/* Get the foreign protocol to use */
    DBGPRINT(DBG$ENTRY,"vms_mail_send: entered",0,0,0,0);
    if((MailProtocolUsed = __transport_name()) == NULL)
      return "No foreign mail transport protocol defined";

/* Create mail context */
    SendContext = 0;
    status = mail$send_begin(&SendContext, NullItemList, NullItemList);
    if ( (err = chksndstat("mail$send_begin", status)) != NULL) return err;

/* Create the sender's address */
/* Use the system default FROM: address (to avoid privileges)
    sprintf(line, "%s <%s@%s>", (from->personal ? from->personal : ""),
	from->mailbox, from->host);
    FromItem[0].length = strlen(line);
    status = mail$send_add_attribute(&SendContext, FromItem, NullItemList);
    if ( (err = chksndstat("mail$send_add_attribute", status)) != NULL) return err;
*/

/* Now the subject */
    if(subject) {
	strcpy(line, subject);
	SubjectItem[0].length = strlen(line);
	status = mail$send_add_attribute(&SendContext, SubjectItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_attribute", status)) != NULL) return err;
    }

/* now add the TO addresses */
    AddressesCount = 0;
    ap = to;
    while(ap != NULL) {
	sprintf(line, "%s%%\"%s@%s\"", MailProtocolUsed, ap->mailbox, ap->host);
	ToItem[0].length = strlen(line);
	status = mail$send_add_address(&SendContext, ToItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_address", status)) != NULL) return err;
	ap = ap->next;
	AddressesCount++;
    }

/* Same for CC */
    ap = cc;
    while(ap != NULL) {
	sprintf(line, "%s%%\"%s@%s\"", MailProtocolUsed, ap->mailbox, ap->host);
	CcItem[0].length = strlen(line);
	status = mail$send_add_address(&SendContext, CcItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_address", status)) != NULL) return err;
	ap = ap->next;
	AddressesCount++;
    }

/* Open the message's file */
    if((ifd = fopen(filename, "r")) == NULL) {
	mail$send_end(&SendContext, NullItemList, NullItemList);
	sprintf(line, "Can't open '%s'", filename);
	return line;
    }

/* Now copy it. VMS/MAIL can accept up to MAXRECORDSIZE characters long lines */

/* --- Method 1: Add each record, one at a time
    while(fgets(line, (sizeof(line) < MAXRECORDSIZE ? sizeof(line) - 1 : MAXRECORDSIZE), ifd) != NULL) {
	char *p = strchr(line, '\n');
	if(p) *p = '\0';
	RecordItem[0].length = strlen(line);
	status = mail$send_add_bodypart(&SendContext, RecordItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_bodypart", status)) != NULL) return err;
    }
*/
    fclose(ifd);

/* --- Method 2: Just add the filename */
    strcpy(line,filename);
    FileItem[0].length = strlen(line);
    status = mail$send_add_bodypart(&SendContext, FileItem, NullItemList);
    if ( (err = chksndstat("mail$send_add_bodypart", status)) != NULL) return err;



/* OK, done - actually send the message. This routine will call via AST a routine
   that will count how many were sent ok.
*/
    MessageSentCount = 0;	/* None sent yet */
    status = mail$send_message(&SendContext, SendItem, NullItemList);
    if ( (err = chksndstat("mail$send_message", status)) != NULL) return err;

    status = mail$send_end(&SendContext, NullItemList, NullItemList);
    if ( (err = chksndstat("mail$send_end", status)) != NULL) return err;

    if(MessageSentCount < AddressesCount)
      {
        switch (MessageSentCount)
          {
            case 0: sprintf(line,"error - no messages sent"); break;
            case 1: sprintf(line,"error - 1 message sent, %d failed",AddressesCount-1); break;
            default:sprintf(line,"error - %d messages sent, %d failed",MessageSentCount, AddressesCount-MessageSentCount); break;
          }
        return line;
      }
    DBGPRINT(DBG$EXIT,"vms_mail_send: exit OK",0,0,0,0);
    return NULL;	/* Success */
}
